import torch
import torch.nn as nn
import torch.optim as optim
from torchvision.models import vgg16_bn, vgg19_bn, vgg13_bn, vgg11_bn, \
                               resnet50, resnet101, resnet152
from pyzjr.Models.bricks.comblock import ConvBNReLU, _DenseBlock

__all__=["FCN8s", "FCN16s", "FCN32s", "FCNs", "VggFCN", "ResFCN", "DenseFCN"]

class FCN32s(nn.Module):
    def __init__(self, model, num_classes):
        super().__init__()
        self.num_classes = num_classes
        self.base_model = model
        self.relu = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1 = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2 = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3 = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4 = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5 = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, num_classes, kernel_size=1)


    def forward(self, x):
        output = self.base_model(x)
        x5 = output["x5"]  # size=(N, 512, x.H/32, x.W/32)

        score = self.bn1(self.relu(self.deconv1(x5)))  # size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)  # size=(N, num_classes, x.H/1, x.W/1)

        return score  # size=(N, num_classes, x.H/1, x.W/1)

class FCN16s(nn.Module):

    def __init__(self, model, num_classes):
        super().__init__()
        self.num_classes = num_classes
        self.base_model = model
        self.relu = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1 = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2 = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3 = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4 = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5 = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, num_classes, kernel_size=1)


    def forward(self, x):
        output = self.base_model(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)
        x4 = output['x4']  # size=(N, 512, x.H/16, x.W/16)

        score = self.relu(self.deconv1(x5))  # size=(N, 512, x.H/16, x.W/16)
        score = self.bn1(score + x4)  # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)  # size=(N, num_classes, x.H/1, x.W/1)

        return score  # size=(N, num_classes, x.H/1, x.W/1)

class FCN8s(nn.Module):
    def __init__(self, model, num_classes):
        super().__init__()
        self.num_classes = num_classes
        self.base_model = model
        self.relu = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1 = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2 = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3 = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4 = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5 = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, num_classes, kernel_size=1)


    def forward(self, x):
        output = self.base_model(x)
        x5 = output["x5"]  # size=(N, 512, x.H/32, x.W/32)
        x4 = output["x4"]  # size=(N, 512, x.H/16, x.W/16)
        x3 = output["x3"]  # size=(N, 256, x.H/8,  x.W/8)
        score = self.relu(self.deconv1(x5))  # size=(N, 512, x.H/16, x.W/16)
        score = self.bn1(score + x4)  # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.relu(self.deconv2(score))  # size=(N, 256, x.H/8, x.W/8)
        score = self.bn2(score + x3)  # element-wise add, size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)  # size=(N, num_classes, x.H/1, x.W/1)

        return score  # size=(N, num_classes, x.H/1, x.W/1)


class FCNs(nn.Module):
    def __init__(self, model, num_classes):
        super().__init__()
        self.num_classes = num_classes
        self.base_model = model
        self.relu = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1 = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2 = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3 = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4 = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5 = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, num_classes, kernel_size=1)

    def forward(self, x):
        output = self.base_model(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)
        x4 = output['x4']  # size=(N, 512, x.H/16, x.W/16)
        x3 = output['x3']  # size=(N, 256, x.H/8,  x.W/8)
        x2 = output['x2']  # size=(N, 128, x.H/4,  x.W/4)
        x1 = output['x1']  # size=(N, 64, x.H/2,  x.W/2)

        score = self.bn1(self.relu(self.deconv1(x5)))  # size=(N, 512, x.H/16, x.W/16)
        score = score + x4  # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = score + x3  # element-wise add, size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = score + x2  # element-wise add, size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = score + x1  # element-wise add, size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)  # size=(N, num_classes, x.H/1, x.W/1)

        return score  # size=(N, num_classes, x.H/1, x.W/1)


class VGGNet(nn.Module):
    def __init__(self, model='vgg16', pretrained=False, requires_grad=True, remove_fc=True, show_params=False):
        ranges = {
            'vgg11': ((0, 4), (4, 8), (8, 15), (15, 22), (22, 29)),
            'vgg13': ((0, 7), (7, 14), (14, 21), (21, 28), (28, 35)),
            'vgg16': ((0, 7), (7, 14), (14, 24), (24, 34), (34, 44)),
            'vgg19': ((0, 7), (7, 14), (14, 27), (27, 40), (40, 53))
        }
        super().__init__()
        if model=="vgg16":
            self.backbone = vgg16_bn(weights='IMAGENET1K_V1' if pretrained else None)
        elif model=="vgg19":
            self.backbone = vgg19_bn(weights='IMAGENET1K_V1' if pretrained else None)
        elif model=="vgg11":
            self.backbone = vgg11_bn(weights='IMAGENET1K_V1' if pretrained else None)
        elif model=="vgg13":
            self.backbone = vgg13_bn(weights='IMAGENET1K_V1' if pretrained else None)
        self.ranges = ranges[model]
        if not requires_grad:
            for param in self.backbone.parameters():
                param.requires_grad = False

        if remove_fc:  # delete redundant fully-connected layer params, can save memory
            del self.backbone.classifier

        if show_params:
            for name, param in self.backbone.named_parameters():
                print(name, param.size())

    def forward(self, x):
        output = {}
        for idx in range(len(self.ranges)):
            for layer in range(self.ranges[idx][0], self.ranges[idx][1]):
                x = self.backbone.features[layer](x)
            output["x%d" % (idx + 1)] = x

        return output

def _get_fun_model(backbone, fcntype, num_classes):
    model = None
    if fcntype == "FCN8s":
        model = FCN8s(model=backbone, num_classes=num_classes)
    elif fcntype == "FCN16s":
        model = FCN16s(model=backbone, num_classes=num_classes)
    elif fcntype == "FCN32s":
        model = FCN32s(model=backbone, num_classes=num_classes)
    elif fcntype == "FCNs":
        model = FCNs(model=backbone, num_classes=num_classes)
    return model

def VggFCN(backbone='vgg16', fcntype="FCN8s", num_classes=1000, pretrained=False):
    """这里仅支持以 vgg_bn 为主干的FCN网络"""
    backbones = VGGNet(model=backbone, pretrained=pretrained)
    model = _get_fun_model(backbones, fcntype, num_classes)
    return model


class ResFCN(nn.Module):
    """这里仅支持以 resnet(50, 101, 152) 为主干的FCN网络"""
    def __init__(self, backbone='resnet152', num_classes=21, pretrained=False, requires_grad=True, remove_fc=True):
        super(ResFCN, self).__init__()
        self.resnet = self._load_resnet(backbone, pretrained, requires_grad, remove_fc)
        self.conv1 = self.resnet.conv1
        self.bn1 = self.resnet.bn1
        self.relu = self.resnet.relu
        self.maxpool = self.resnet.maxpool
        self.layer1 = self.resnet.layer1
        self.layer2 = self.resnet.layer2
        self.layer3 = self.resnet.layer3
        self.layer4 = self.resnet.layer4
        # Decoder
        self.decoder1 = nn.Sequential(ConvBNReLU(2048, num_classes, kernel_size=1, padding=0, bias=True),
            nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1))
        self.layer3_conv = ConvBNReLU(1024, num_classes, kernel_size=1, padding=0, bias=True)
        self.decoder2 = nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1)
        self.layer2_conv = ConvBNReLU(512, num_classes, kernel_size=1, padding=0, bias=True)
        self.decoder3 = nn.ConvTranspose2d(num_classes, num_classes, kernel_size=16, stride=8, padding=4)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x = self.layer4(x3)
        x = self.decoder1(x) + self.layer3_conv(x3)
        x = self.decoder2(x) + self.layer2_conv(x2)
        x = self.decoder3(x)

        return x

    def _load_resnet(self, model_type, pretrained, requires_grad, remove_fc):
        if model_type == "resnet50":
            self.resnet = resnet50(weights='IMAGENET1K_V1' if pretrained else None)
        elif model_type == "resnet101":
            self.resnet = resnet101(weights='IMAGENET1K_V1' if pretrained else None)
        elif model_type == "resnet152":
            self.resnet = resnet152(weights='IMAGENET1K_V1' if pretrained else None)

        if not requires_grad:
            for param in self.resnet.parameters():
                param.requires_grad = False

        if remove_fc:
            del self.resnet.fc

        return self.resnet


class DenseFCN(nn.Module):
    def __init__(self, num_classes=21):
        super(DenseFCN, self).__init__()
        # First convolution
        self.conv1 = ConvBNReLU(3, 64, bias=True)
        self.maxpool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.DB1 = _DenseBlock(6, 64, 4, 32, 0)
        self.conv2 = ConvBNReLU(256, 128, kernel_size=1, padding=0, bias=True)
        self.DB2 = _DenseBlock(12, 128, 4, 32, 0)
        self.conv3 = ConvBNReLU(512, 256, kernel_size=1, padding=0, bias=True)
        self.DB3 = _DenseBlock(24, 256, 4, 32, 0)
        self.conv4 = ConvBNReLU(1024, 512, kernel_size=1, padding=0, bias=True)
        self.DB4 = _DenseBlock(16, 512, 4, 32, 0)
        self.conv5 = nn.Sequential(ConvBNReLU(1024, 1000, bias=True),
                                   ConvBNReLU(1000, num_classes, kernel_size=1, padding=0, bias=True))
        self.deconv1 = nn.Sequential(nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1),
                                     ConvBNReLU(num_classes, num_classes, kernel_size=1, padding=0, bias=True))
        self.deconv2 = nn.Sequential(nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1),
                                     ConvBNReLU(num_classes, num_classes, kernel_size=1, padding=0, bias=True))
        self.deconv3 = nn.Sequential(nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1),
                                     ConvBNReLU(num_classes, num_classes, kernel_size=1, padding=0, bias=True))
        self.deconv4 = nn.Sequential(nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1),
                                     ConvBNReLU(num_classes, num_classes, kernel_size=1, padding=0, bias=True))
        self.deconv5 = nn.Sequential(nn.ConvTranspose2d(num_classes, num_classes, kernel_size=4, stride=2, padding=1),
                                     ConvBNReLU(num_classes, num_classes, kernel_size=1, padding=0, bias=True))

    def forward(self, x):
        x = self.conv1(x)
        x = self.maxpool(x)
        x = self.DB1(x)
        x = self.conv2(x)
        x = self.maxpool(x)
        x = self.DB2(x)
        x = self.conv3(x)
        x = self.maxpool(x)
        x = self.DB3(x)
        x = self.conv4(x)
        x = self.maxpool(x)
        x = self.DB4(x)
        x = self.maxpool(x)
        x = self.conv5(x)

        x = self.deconv1(x)
        x = self.deconv2(x)
        x = self.deconv3(x)
        x = self.deconv4(x)
        x = self.deconv5(x)

        return x







if __name__ == "__main__":
    batch_size, num_classes, h, w = 10, 21, 160, 160
    fcn_model = DenseFCN(num_classes=num_classes)
    # fcn_model = ResFCN(num_classes=num_classes)
    # fcn_model = VggFCN(backbone='vgg16', fcntype="FCN8s", num_classes=num_classes)
    criterion = nn.BCELoss()
    optimizer = optim.SGD(fcn_model.parameters(), lr=1e-3)
    input = torch.randn(batch_size, 3, h, w, requires_grad=True)
    y = torch.randn(batch_size, num_classes, h, w, requires_grad=False)
    # 损失应当下降
    for iter in range(10):
        optimizer.zero_grad()
        output = fcn_model(input)
        output = torch.sigmoid(output)
        loss = criterion(output, y)
        loss.backward()
        print(f"epoch {iter}, loss {loss.item()}")
        print(output.shape)
        optimizer.step()


# 预训练模型地址

from torch.hub import load_state_dict_from_url

model_urls = {
    'vgg16_bn': 'https://download.pytorch.org/models/vgg16_bn-6c64b313.pth',
    'vgg19_bn': 'https://download.pytorch.org/models/vgg19_bn-c79401a0.pth',
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
    'alexnet': 'https://download.pytorch.org/models/alexnet-owt-4df8aa71.pth',
    'se_resnet50': 'https://github.com/moskomule/senet.pytorch/releases/download/archive/seresnet50-60a8950a85b2b.pkl',
    'googlenet': 'https://download.pytorch.org/models/googlenet-1378be20.pth',
}

def download_from_url(name="resnet50", model_dir=None, progress=True):
    if model_dir is None:
        model_dir = './model_weight'
    load_state_dict_from_url(model_urls[name], model_dir=model_dir, progress=progress)

if __name__=="__main__":
    name = "googlenet"

    model_dir = "D:\PythonProject\Crack_classification_training_script\model_weight"
    download_from_url(name=name, model_dir=model_dir)


import os
import re
from pathlib import Path
import random
from tqdm import tqdm
import cv2
import shutil
import numpy as np
import PIL.Image as Image
import datetime
from os import getcwd

__all__ = ["getPhotopath", "timestr", "pathstr", "natsorted", "natural_key",
           "MultiImageFolderProcessor"]

IMAGE_EXTENSIONS = ('.png', '.jpg', '.jpeg', '.gif', '.bmp', '.tiff', '.webp', 'tif', 'raw')
IMAGE_EXTENSIONS_SET = set(IMAGE_EXTENSIONS)

def getPhotopath(paths, cd=False, debug=False):
    """
    :param paths: 文件夹路径
    :param cd: 添加当前运行的路径名,这是使用了相对路径才能用的
    :param debug: 开启打印文件名错误的名字
    :return: 包含图片路径的列表
    """
    imgfile = []
    allfile = []
    file_list = os.listdir(paths)
    for i in file_list:
        if debug:
            if i[0] in ['n', 't', 'r', 'b', 'f'] or i[0].isdigit():
                print(f"[pyzjr]: File name:Error occurred at the beginning of {i}!")
        newph = os.path.join(paths, i).replace("\\", "/")
        allfile.append(newph)
        _, file_ext = os.path.splitext(newph)
        if file_ext[1:] in IMAGE_EXTENSIONS:
            imgfile.append(newph)
    if cd:
        # 谨慎使用
        cdd = getcwd()
        imgfile = [os.path.join(cdd, file).replace("\\", "/") for file in imgfile]
        allfile = [os.path.join(cdd, file).replace("\\", "/") for file in allfile]
    return imgfile, allfile

def timestr():
    """Generate a formatted datetime string."""
    return f"{datetime.datetime.now():%Y_%m_%d_%H_%M_%S}"

def pathstr(path_str):
    """
    path_str = 'D:/PythonProject/Torchproject/Lasercenterline/line/20231013-LaserLine_txt/test_2/imges/Image_1.jpg'
    D:\
    PythonProject
    Torchproject
    Lasercenterline
    line
    20231013-LaserLine_txt
    test_2
    imges
    Image_1.jpg
    """
    path = Path(path_str)
    path_parts = path.parts
    return list(path_parts)

def natural_key(st):
    """
    将字符串拆分成字母和数字块
    """
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split('([0-9]+)', st)]

def natsorted(a):
    """
    手写自然排序
    >>> a = ['num9', 'num5', 'num2']
    >>> sorted_a = natsorted(a)
    ['num2', 'num5', 'num9']
    """
    return sorted(a, key=natural_key)

class MultiImageFolderProcessor():
    def __init__(self, imagefolder):
        if self.is_directory_not_empty(imagefolder):
            self.imagefolder = imagefolder
        else:
            raise ValueError("[MultiImageFolderProcessor]:The provided folder path is empty.")

    def is_image_file(self, filename):
        """
        检查文件名是否符合要求
        :param filename: 文件名
        """
        return any(filename.lower().endswith(ext) for ext in IMAGE_EXTENSIONS)

    def is_image_extension(self, ext):
        """
        检查给定的文件扩展名是否为图像扩展名。
        :param ext (str): 要检查的文件扩展名。
        """
        return ext in IMAGE_EXTENSIONS_SET

    def valid_extension(self, ext):
        """
        检查给定的文件扩展名是否是有效的扩展名。
        :param ext (str): 要检查的文件扩展名。
        """
        if ext is not None and isinstance(ext, str) and ext.startswith('.') and len(ext) >= 2:
            return True
        else:
            return False

    def is_directory_not_empty(self, path):
        """
        检查给定路径是否是非空文件夹。
        :param path (str): 要检查的路径。
        """
        return os.path.isdir(path) and len(os.listdir(path)) > 0

    @property
    def list_dirs(self):
        """
        返回给定目录中的所有目录
        """
        return [f for f in Path(self.imagefolder).iterdir() if f.is_dir()]

    @property
    def list_files(self):
        """
        返回给定目录中的所有文件
        """
        return [
            f for f in Path(self.imagefolder).iterdir()
            if f.is_file() and not f.name.startswith(".")
        ]

    @property
    def run_py_path(self):
        """当前运行的py脚本的路径"""
        return os.path.abspath(__file__)

    @property
    def run_py_dir_path(self):
        """当前运行py脚本的文件夹路径"""
        return os.path.dirname(os.path.abspath(__file__))

    def list_files_by_ext(self, ext):
        """
        返回给定目录中指定扩展名的文件列表
        :param ext(str): 指定的文件扩展名。
        """
        return [
            f for f in Path(self.imagefolder).iterdir()
            if f.is_file() and not f.name.startswith(".") and f.suffix == ext
        ]

    @property
    def get_imagepath_list(self):
        """获取符合IMAGE_EXTENSIONS的图像路径"""
        imgpathlist = [
            f for f in Path(self.imagefolder).iterdir()
            if f.is_file() and not f.name.startswith(".") and f.suffix in IMAGE_EXTENSIONS
        ]
        if len(imgpathlist)==0:
            raise ValueError(f"[MultiImageFolderProcessor]:The obtained image list is empty, possibly because the suffix does not match {IMAGE_EXTENSIONS}")
        return imgpathlist

    def split_seg_trainval_txt(self, save_folder, tratio=0.8, vratio=0.1, shuffle=True):
        """
        划分分割任务的训练集、测试集、验证集的脚本
        :param save_folder: 指定的文件保存路径
        :param tratio: 训练集比例
        :param vratio: 验证集比例
        :param shuffle: 打乱写入顺序
        """
        def write_txt(file_list, txt_path, desc):
            total_files = len(file_list)
            with tqdm(total=total_files, desc=desc, unit=" per") as pbar:
                with open(txt_path, 'w') as f:
                    for file_name in file_list:
                        f.write(file_name[:-4] + '\n')
                        pbar.update(1)
        os.makedirs(save_folder, exist_ok=True)

        test_ratio = 1 - tratio - vratio
        if test_ratio < 0 or test_ratio > 1:
            raise ValueError(f"[MultiImageFolderProcessor]:The ratio between tratio={tratio} and vratio={vratio} is incorrect, and the sum of the two should be less than or equal to 1")
        jpg_files = [file for file in os.listdir(self.imagefolder) if file.lower().endswith('.jpg')]
        if len(jpg_files) == 0:
            raise ValueError(f"[MultiImageFolderProcessor]:The {self.imagefolder} has been converted to an empty list, which may be due to images not being in JPG format.")
        if shuffle:
            random.shuffle(jpg_files)
        total_files = len(jpg_files)
        train_files = int(tratio * total_files)
        val_files = int(vratio * total_files)

        train_list = jpg_files[:train_files]
        val_list = jpg_files[train_files:train_files + val_files]
        test_list = jpg_files[train_files + val_files:] if test_ratio > 0 else []
        write_txt(train_list, os.path.join(save_folder, 'train.txt'),desc="Write training file")
        write_txt(val_list, os.path.join(save_folder, 'val.txt'), desc="Write verification file")
        write_txt(test_list, os.path.join(save_folder, 'test.txt'), desc="Write testing file")

    def convert_image_format(self, save_folder, output_format='.png'):
        """
        任意格式转换
        :param save_folder: 指定的文件保存路径
        :param output_format: 指定修改的后缀名
        """
        os.makedirs(save_folder, exist_ok=True)
        image_files = [file for file in os.listdir(self.imagefolder) if file.lower().endswith(IMAGE_EXTENSIONS)]

        for image_file in image_files:
            input_path = os.path.join(self.imagefolder, image_file)
            output_path = os.path.join(save_folder, os.path.splitext(image_file)[0] + output_format)

            img = cv2.imread(input_path)
            cv2.imwrite(output_path, img)

    def copy_image_folder(self, save_folder):
        """
        复制整个文件夹（图像）到另外一个文件夹
        :param save_folder: 指定的文件保存路径
        """
        try:
            os.makedirs(save_folder, exist_ok=True)
            for root, dirs, files in os.walk(self.imagefolder):
                for file in files:
                    if self.is_image_file(file):
                        source_path = os.path.join(root, file)
                        destination_path = os.path.join(save_folder, file)  # Fix: use a different variable for destination path
                        shutil.copy2(source_path, destination_path)
            print(f"[MultiImageFolderProcessor]:Successfully copied folder: {self.imagefolder} to {save_folder}")
        except Exception as e:
            print(f"[MultiImageFolderProcessor]:Error copying folder: {e}")

    def resize_image_folder(self, save_folder, target_size):
        """
        修改图像的大小，并保存到指定的文件夹
        :param save_folder: 指定的文件保存路径
        :param target_size: 指定的图像大小，可以是一个整数，也可以是长度为2的列表或元组
        """
        def check_image_size(target_size):
            if isinstance(target_size, int):
                target_size = (target_size, target_size)
                return target_size
            if isinstance(target_size, (list, tuple)) and len(target_size) == 2:
                return target_size
            else:
                raise ValueError("[MultiImageFolderProcessor]:Invalid target_size format. Please provide a single integer or a tuple/list of two integers.")
        try:
            target_size = check_image_size(target_size)
            os.makedirs(save_folder, exist_ok=True)
            for root, dirs, files in os.walk(self.imagefolder):
                for file in files:
                    if file.lower().endswith(IMAGE_EXTENSIONS):
                        source_path = os.path.join(root, file)
                        output_path = os.path.join(save_folder, file)
                        img = cv2.imread(source_path)
                        resized_img = cv2.resize(img, target_size)
                        cv2.imwrite(output_path, resized_img)

            print(f"[MultiImageFolderProcessor]:Successfully resized and saved images from {self.imagefolder} to {save_folder}")
        except Exception as e:
            print(f"[MultiImageFolderProcessor]:Error resizing and saving images: {e}")

    def check_2_folder_different_names(self, png_folder, save_folder):
        """
        检查两个文件夹当中图片名是否相同,如果不同,移动这些文件到新的文件夹当中(不包含后缀的情况,一般用于检查训练集和测试集),这样就能划分出测试集
        :param target_folder:
        :param save_folder:
        """
        os.makedirs(save_folder, exist_ok=True)
        # 获取两个文件夹中的文件名称（不包括文件扩展名）
        jpg_files = {os.path.splitext(file)[0] for file in os.listdir(self.imagefolder) if file.lower().endswith('.jpg')}
        png_files = {os.path.splitext(file)[0] for file in os.listdir(png_folder) if file.lower().endswith('.png')}
        if len(jpg_files) == 0:
            raise ValueError(f"[MultiImageFolderProcessor]:The file of {self. imagefolder} is recognized, possibly because it does not end in jpg format")
        if len(png_files) == 0:
            raise ValueError(f"[MultiImageFolderProcessor]:The file of {png_folder} is recognized, possibly because it does not end in png format or the file path is incorrect")
        if len(jpg_files) >= len(png_files):
            unmatched_files = jpg_files - png_files
            flag_to_move_jpg_files = True
        else:
            unmatched_files = png_files - jpg_files
            flag_to_move_jpg_files = False

        for file in unmatched_files:
            if flag_to_move_jpg_files:
                shutil.move(os.path.join(self.imagefolder, file + '.jpg'), os.path.join(save_folder, file + '.jpg'))
            else:
                shutil.move(os.path.join(png_folder, file + '.png'), os.path.join(save_folder, file + '.png'))

    @property
    def get_mean_std(self):
        """
        :return: 返回RGB顺序的mean与std
        """
        all_images = [Image.open(os.path.join(self.imagefolder, file)).convert('RGB') for file in
                      os.listdir(self.imagefolder) if file.lower().endswith(IMAGE_EXTENSIONS)]
        num_images = len(all_images)
        mean_sum = [0, 0, 0]
        std_sum = [0, 0, 0]

        for image in all_images:
            img_asarray = np.asarray(image) / 255.0
            individual_mean = np.mean(img_asarray, axis=(0, 1))
            individual_stdev = np.std(img_asarray, axis=(0, 1))
            mean_sum += individual_mean
            std_sum += individual_stdev

        mean = mean_sum / num_images
        std = std_sum / num_images
        return list(mean), list(std)

    def RenameImageFolder(self, save_folder, newbasename='img', type=3, format=None):
        """
        重命名图像文件夹中的所有图像文件
        :param save_folder: 新文件夹的保存路径
        :param type: 数字长度，比如005长度为3
        :param format: 后缀,如果是None就表示用原来的后缀
        """
        os.makedirs(save_folder, exist_ok=True)
        imglist = self.get_imagepath_list
        total_files = len(imglist)
        for i, file in tqdm(enumerate(imglist), total=total_files, desc='Renaming files'):
            properties = os.path.basename(file)
            name, ext = os.path.splitext(properties)
            padded_i = str(i).zfill(type)
            if format is not None:
                newname = f"{newbasename}{padded_i}." + format
            else:
                newname = f"{newbasename}{padded_i}.{ext[1:]}"
            new_path = os.path.join(save_folder, newname)
            shutil.copy(file, new_path)
        print("[MultiImageFolderProcessor]:Batch renaming and saving of files completed!")



if __name__=="__main__":
    folder = r"D:\PythonProject\pythonProject2\dataloader\Pothole\test\images"
    folderimg = MultiImageFolderProcessor(folder)
    # folderimg.split_seg_trainval_txt(save_folder=r"D:\PythonProject\pythonProject2\dataloader\Pothole")
    # a = folderimg.list_files
    # folderimg.resize_image_folder(save_folder=r"D:\PythonProject\pythonProject2\dataloader\Pothole\test3", target_size=512)
    # folderimg.RenameImageFolder(save_folder=r"D:\PythonProject\pythonProject2\dataloader\Pothole\test3")





import torch
import torch.nn as nn

class DropPath(nn.Module):
    """
    Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).

    I follow the implementation:
        https://github.com/rwightman/pytorch-image-models/blob/a2727c1bf78ba0d7b5727f5f95e37fb7f8866b1f/timm/models/layers/drop.py
    """
    def __init__(self, drop_prob = 0.):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        """
        DropPath类实现了"Drop Path"的正则化技术，用于深度神经网络中的主路径，特别是在残差块的主路径上。Drop Path类似于Dropout，
        但它是在网络的深度路径上随机丢弃整个路径（或称之为跳跃连接）。这有助于训练更加健壮和泛化性能更好的神经网络。
        """
        return self.drop_path(x, self.drop_prob, self.training)

    def __str__(self):
        return f'drop_prob={round(self.drop_prob,3):0.3f}'

    def drop_path(self, x, drop_prob = 0., training = False):
        """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
        This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
        the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
        See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
        changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
        'survival rate' as the argument.
        """
        if drop_prob == 0. or not training:
            return x
        keep_prob = 1 - drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
        random_tensor = keep_prob + torch.rand(
            shape, dtype=x.dtype, device=x.device)
        output = x.div(keep_prob) * random_tensor.floor()
        return output



"""
Paper address: <https://arxiv.org/pdf/2303.03667.pdf>
Reference from: https://github.com/JierunChen/FasterNet/blob/master/models/fasternet.py
Blog records: https://blog.csdn.net/m0_62919535/article/details/136334105
"""
import torch
import torch.nn as nn
from pyzjr.Models.bricks import DropPath

__all__=["FasterNet", "FasterNetBlock", "fasternet_t0", "fasternet_t1", "fasternet_t2",
         "fasternet_s", "fasternet_m", "fasternet_l"]

class PartialConv(nn.Module):
    def __init__(self, dim, n_div=4, kernel_size=3, forward='split_cat'):
        """
        PartialConv 模块

        Args:
            dim (int): 输入张量的通道数。
            n_div (int): 分割通道数的分母，用于确定部分卷积的通道数。
            forward (str): 使用的前向传播方法，可选 'slicing' 或 'split_cat'。
        """
        super().__init__()
        self.dim_conv3 = dim // n_div
        self.dim_untouched = dim - self.dim_conv3
        self.partial_conv3 = nn.Conv2d(self.dim_conv3, self.dim_conv3, kernel_size, 1, 1, bias=False)

        if forward == 'slicing':
            self.forward = self.forward_slicing
        elif forward == 'split_cat':
            self.forward = self.forward_split_cat
        else:
            raise NotImplementedError

    def forward_slicing(self, x):
        # only for inference
        x = x.clone()   # !!! Keep the original input intact for the residual connection later
        x[:, :self.dim_conv3, :, :] = self.partial_conv3(x[:, :self.dim_conv3, :, :])

        return x

    def forward_split_cat(self, x):
        # for training/inference
        x1, x2 = torch.split(x, [self.dim_conv3, self.dim_untouched], dim=1)
        x1 = self.partial_conv3(x1)
        x = torch.cat((x1, x2), 1)

        return x


class FasterNetBlock(nn.Module):
    def __init__(self, dim, expand_ratio=2, act_layer=nn.ReLU, drop_path_rate=0.0, forward='split_cat'):
        super().__init__()
        self.pconv = PartialConv(dim, forward=forward)
        self.conv1 = nn.Conv2d(dim, dim * expand_ratio, 1, bias=False)
        self.bn = nn.BatchNorm2d(dim * expand_ratio)
        self.act_layer = act_layer()
        self.conv2 = nn.Conv2d(dim * expand_ratio, dim, 1, bias=False)
        self.drop_path = DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()

    def forward(self, x):
        residual = x
        x = self.pconv(x)
        x = self.conv1(x)
        x = self.bn(x)
        x = self.act_layer(x)
        x = self.conv2(x)
        x = residual + self.drop_path(x)
        return x


class FasterNet(nn.Module):
    def __init__(self, in_channel=3, embed_dim=40, act_layer=None,
                 num_classes=1000, depths=None, drop_rate=0.0):
        super().__init__()
        # Embedding
        self.stem = nn.Sequential(
            nn.Conv2d(in_channel, embed_dim, 4, stride=4, bias=False),
            nn.BatchNorm2d(embed_dim),
            act_layer()
        )
        drop_path_list = [x.item() for x in torch.linspace(0, drop_rate, sum(depths))]
        self.feature = []
        embed_dim = embed_dim
        for idx, depth in enumerate(depths):
            self.feature.append(nn.Sequential(
                *[FasterNetBlock(embed_dim, act_layer=act_layer, drop_path_rate=drop_path_list[sum(depths[:idx]) + i]) for i in range(depth)]
            ))
            if idx < len(depths) - 1:
                # Merging
                self.feature.append(nn.Sequential(
                    nn.Conv2d(embed_dim, embed_dim * 2, 2, stride=2, bias=False),
                    nn.BatchNorm2d(embed_dim * 2),
                    act_layer()
                ))
                embed_dim = embed_dim * 2

        self.feature = nn.Sequential(*self.feature)
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.conv1 = nn.Conv2d(embed_dim, 1280, 1, bias=False)
        self.act_layer = act_layer()
        self.fc = nn.Linear(1280, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.stem(x)
        x = self.feature(x)
        x = self.avg_pool(x)
        x = self.conv1(x)
        x = self.act_layer(x)
        x = self.fc(x.flatten(1))

        return x


def fasternet_t0(num_classes, drop_path_rate=0.0):
    return FasterNet(embed_dim=40,
                     act_layer=nn.GELU,
                     num_classes=num_classes,
                     depths=[1, 2, 8, 2],
                     drop_rate=drop_path_rate
                     )

def fasternet_t1(num_classes, drop_path_rate=0.02):
    return FasterNet(embed_dim=64,
                     act_layer=nn.GELU,
                     num_classes=num_classes,
                     depths=[1, 2, 8, 2],
                     drop_rate=drop_path_rate
                     )

def fasternet_t2(num_classes, drop_path_rate = 0.05):
    return FasterNet(embed_dim=96,
                     act_layer=nn.ReLU,
                     num_classes=num_classes,
                     depths=[1, 2, 8, 2],
                     drop_rate=drop_path_rate
                     )

def fasternet_s(num_classes, drop_path_rate = 0.03):
    return FasterNet(embed_dim=128,
                     act_layer=nn.ReLU,
                     num_classes=num_classes,
                     depths=[1, 2, 13, 2],
                     drop_rate=drop_path_rate
                     )

def fasternet_m(num_classes, drop_path_rate = 0.05):
    return FasterNet(embed_dim=144,
                     act_layer=nn.ReLU,
                     num_classes=num_classes,
                     depths=[3, 4, 18, 3],
                     drop_rate=drop_path_rate
                     )

def fasternet_l(num_classes, drop_path_rate = 0.05):
    return FasterNet(embed_dim=192,
                     act_layer=nn.ReLU,
                     num_classes=num_classes,
                     depths=[3, 4, 18, 3],
                     drop_rate=drop_path_rate
                     )

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = fasternet_t2(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # t0 Total params: 2,629,624 Estimated Total Size (MB): 57.70
    # t1 Total params: 6,319,492 Estimated Total Size (MB): 100.02
    # t2 Total params: 13,707,012 Estimated Total Size (MB): 165.87
    # s Total params: 29,905,156 Estimated Total Size (MB): 304.56
    # m Total params: 52,245,588 Estimated Total Size (MB): 568.01
    # l Total params: 92,189,316 Estimated Total Size (MB): 843.09


import numpy as np
from PIL import Image
import torch
import torch.nn as nn

__all__ = ["is_numpy","is_pil","is_tensor","is_rgb_image","is_gray_image","get_num_channels","get_image_size",
           "is_parallel"]


def is_tensor(img):
    """Check if the input image is torch format."""
    return isinstance(img, torch.Tensor)

def is_pil(img):
    """Check if the input image is PIL format."""
    return isinstance(img, Image.Image)

def is_numpy(img):
    """Check if the input image is Numpy format."""
    return isinstance(img, np.ndarray)

def is_gray_image(image):
    return (len(image.shape) == 2) or (len(image.shape) == 3 and image.shape[-1] == 1)

def is_rgb_image(image):
    return len(image.shape) == 3 and image.shape[-1] == 3

def get_num_channels(image):
    return image.shape[2] if len(image.shape) == 3 else 1

def get_image_size(image):
    if is_numpy(image):
        h, w = image.shape[:2]
        return h, w
    if is_pil(image):
        w, h = image.size
        return h, w
    if is_tensor(image):
        if len(image.shape) == 4 or len(image.shape) == 3:
            w, h = image.shape[-2:]
            return h, w
    else:
        raise ValueError("[pyzjr]:Unsupported input type")

def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))





"""
这一部分主要针对于opencv格式的,一般这部分用于传统图像的增强
"""
import numpy as np
import cv2
from math import ceil
import random
from pyzjr.core import _check_parameter_is_tuple_2, _check_parameter_is_tuple_and_list_2

__all__ = ["base_crop1", "base_crop2", "center_crop", "five_crop", "Stitcher_image",\

           "Centerzoom", "flip", "horizontal_flip", "vertical_flip", "resize", "adjust_brightness_cv2", "adjust_brightness_numpy",\

           "rotate", "adjust_gamma", "pad", "erase", "augment_Hsv","hist_equalize","random_resize_crop","random_crop",\

           "random_horizontal_flip", "random_vertical_flip", "random_rotation", "random_lighting", "random_apply", "random_order",\

           "random_choice", "uniform_augment",

           "Retinex", "blur", "median_blur", "gaussian_blur", "bilateral_blur", "Filter"]

def base_crop1(img, x_min, y_min, x_max, y_max):
    height, width = img.shape[:2]
    assert x_max > x_min, "[pyzjr]:Maximum value of cropping x_max cannot be less than the minimum value x_min"
    assert y_max > y_min, "[pyzjr]:Maximum value of cropping y_max cannot be less than the minimum value y_min"
    assert x_min >= 0, "[pyzjr]:x_min cannot be less than 0"
    assert x_max <= width, "[pyzjr]:x_max cannot be greater than the image width"
    assert y_min >= 0, "[pyzjr]:y_min cannot be less than 0"
    assert y_max <= height, "[pyzjr]:y_max cannot be greater than the image height"

    return img[y_min:y_max, x_min:x_max]

def base_crop2(img, x_start, y_start, width, height):
    assert width > 0 and height > 0, "[pyzjr]:Width and height of cropping area must be greater than 0"
    assert x_start >= 0 and y_start >= 0, "[pyzjr]:x_min and y_min cannot be less than 0"

    return img[y_start:y_start+height, x_start:x_start+width]

def center_crop(image, target_size):
    """
    Center-crops an image to the specified target size.

    Args:
        image (numpy.ndarray): The input image.
        target_size (tuple): A tuple (width, height) specifying the target size.

    Returns:
        numpy.ndarray: The center-cropped image.
    """
    h, w = image.shape[:2]
    crop_width, crop_height = target_size

    if crop_width > w or crop_height > h:
        raise ValueError("[pyzjr]:Target size is larger than the input image size")

    x_start = (w - crop_width) // 2
    y_start = (h - crop_height) // 2

    return base_crop2(image,x_start,y_start,crop_width,crop_height)

def five_crop(image, size):
    """
    Generate 5 cropped images (one central and four corners).

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        size (tuple): A tuple (width, height) specifying the crop size.

    Returns:
        list: A list of 5 NumPy arrays.
    """
    _check_parameter_is_tuple_2(size)

    width, height = image.shape[1], image.shape[0]
    crop_width, crop_height = size

    if crop_width > width or crop_height > height:
        raise ValueError("Crop size exceeds the image dimensions.")

    center_x = width // 2
    center_y = height // 2

    crops = []

    # Central crop
    left = center_x - crop_width // 2
    upper = center_y - crop_height // 2
    right = left + crop_width
    lower = upper + crop_height
    central_crop = image[upper:lower, left:right]
    crops.append(central_crop)

    # Top-left corner crop
    top_left_crop = image[0:crop_height, 0:crop_width]
    crops.append(top_left_crop)

    # Top-right corner crop
    top_right_crop = image[0:crop_height, width - crop_width:width]
    crops.append(top_right_crop)

    # Bottom-left corner crop
    bottom_left_crop = image[height - crop_height:height, 0:crop_width]
    crops.append(bottom_left_crop)

    # Bottom-right corner crop
    bottom_right_crop = image[height - crop_height:height, width - crop_width:width]
    crops.append(bottom_right_crop)

    return crops    # central_crop, top_left_crop, top_right_crop, bottom_left_crop, bottom_right_crop

def Stitcher_image(image_paths):
    """
    图像拼接，图片较小可能拼接失败
    :param image_paths: 由图片路径组成的列表
    :return: 返回被拼接好的图片
    """
    opencv_version = cv2.__version__
    major_version = int(opencv_version.split(".")[0])
    if major_version >= 4:
        stitcher = cv2.Stitcher.create() # 适用于OpenCV 4.x
    else:
        stitcher = cv2.createStitcher()  # 适用于OpenCV 3.x 或更早版本
    images = []
    for path in image_paths:
        img = cv2.imread(path)
        if img is not None:
            images.append(img)
    assert len(images) >= 2, "[pyzjr]:At least two images are required for stitching"
    (status, stitched_image) = stitcher.stitch(images)
    assert status == cv2.Stitcher_OK, '[pyzjr]:Image stitching failed'
    return stitched_image

def Centerzoom(img, zoom_factor: int):
    """中心缩放"""
    h, w = img.shape[:2]
    h_ch, w_ch = ceil(h / zoom_factor), ceil(w / zoom_factor)
    h_top, w_top = (h - h_ch) // 2, (w - w_ch) // 2
    zoomed_img = cv2.resize(img[h_top : h_top + h_ch, w_top : w_top + w_ch], (w, h), interpolation=cv2.INTER_LINEAR)

    return zoomed_img


def flip(image, option_value):
    """
    :param image : numpy array of image
    :param option_value: random integer between 0 to 3
            vertical                          0
            horizontal                        1
            horizontally and vertically flip  2
    Return: image : numpy array of flipped image
    """
    if option_value == 0:
        image = np.flip(image, option_value)
    elif option_value == 1:
        image = np.flip(image, option_value)
    elif option_value == 2:
        image = np.flip(image, 0)
        image = np.flip(image, 1)
    else:
        image = image

    return image

def horizontal_flip(img):
    """Flip the image horizontally"""
    return flip(img, 1)

def vertical_flip(img):
    """Flip the image vertically"""
    return flip(img, 0)

def resize(image, size=None, scale=None):
    """
    双线性插值 https://blog.csdn.net/m0_62919535/article/details/132094815
    Args:
        image (numpy.ndarray): Image to be resized.
        size (tuple): New size in the format (width, height).
        scale (float): Scale of image
    Returns:
        Resized image.
    """
    if scale is not None:
        ah, aw, channel = image.shape
        bh, bw = int(ah * scale), int(aw * scale)
        dst_img = np.zeros((bh, bw, channel), np.uint8)

        y_coords, x_coords = np.meshgrid(np.arange(bh), np.arange(bw), indexing='ij')
        AX = (x_coords + 0.5) / scale - 0.5
        AY = (y_coords + 0.5) / scale - 0.5

        x1 = np.floor(AX).astype(int)
        y1 = np.floor(AY).astype(int)
        x2 = np.minimum(x1 + 1, aw - 1)
        y2 = np.minimum(y1 + 1, ah - 1)
        R1 = ((x2 - AX)[:, :, np.newaxis] * image[y1, x1]).astype(float) + (
                (AX - x1)[:, :, np.newaxis] * image[y1, x2]).astype(float)
        R2 = ((x2 - AX)[:, :, np.newaxis] * image[y2, x1]).astype(float) + (
                (AX - x1)[:, :, np.newaxis] * image[y2, x2]).astype(float)

        dst_img = (y2 - AY)[:, :, np.newaxis] * R1 + (AY - y1)[:, :, np.newaxis] * R2

    if size is not None:
        dst_img = cv2.resize(image, size, interpolation=cv2.INTER_LINEAR)

    return dst_img.astype(np.uint8)

def adjust_brightness_cv2(image, brightness_factor):
    """
    Adjust brightness of an image using OpenCV.
    Args:
        image (numpy.ndarray): Image to be adjusted.
        brightness_factor (float): A factor by which to adjust brightness.
            - 0.0 gives a black image.
            - 1.0 gives the original image.
            - Greater than 1.0 increases brightness.
            - Less than 1.0 decreases brightness.
    Returns:
        Brightness-adjusted image.
    """
    return cv2.convertScaleAbs(image, alpha=brightness_factor, beta=0)

def adjust_brightness_numpy(image, brightness_factor):
    """
    Adjust brightness of an image using Numpy.
    Args:
        image (numpy.ndarray): Image to be adjusted.
        brightness_factor (float): A factor by which to adjust brightness.
            - 0.0 gives a black image.
            - 1.0 gives the original image.
            - Greater than 1.0 increases brightness.
            - Less than 1.0 decreases brightness.
    Returns:
        Brightness-adjusted image.
    """
    image_float = image.astype(np.float32)
    _image = image_float * brightness_factor
    _image = np.clip(_image, 0, 255)
    b_image = _image.astype(np.uint8)
    return b_image

def rotate(image, angle, fill_value=(0, 0, 0)):
    """
    Rotate the input image by angle degrees.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        angle (float): Rotation angle in degrees, counter-clockwise.
        fill_value (tuple, optional): Fill color for areas outside the rotated image.
            Default is (0, 0, 0) for black.

    Returns:
        numpy.ndarray: Rotated image.

    """
    height, width = image.shape[:2]
    rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), angle, 1)
    rotated_image = cv2.warpAffine(image, rotation_matrix, (width, height), borderValue=fill_value)
    return rotated_image

def adjust_gamma(img, gamma, gain=1):
    """Adjust gamma of the input opencv image."""
    img_np = np.array(img)
    img_gamma_corrected = ((img_np / 255.0) ** gamma) * 255.0 * gain
    img_gamma_corrected = np.clip(img_gamma_corrected, 0, 255).astype(np.uint8)
    return img_gamma_corrected

def pad(img, padding, fill_value=(128, 128, 128)):
    if not isinstance(padding, (int, tuple, list)):
        raise TypeError("Padding should be an integer or a tuple/list of two or four values.")

    if not isinstance(fill_value, (int, str, tuple)):
        raise TypeError("Fill value should be an integer, a string, or a tuple.")

    top = bottom = left = right = None
    if isinstance(padding, int):
        top = bottom = left = right = padding
    elif len(padding) == 2:
        left = right = padding[0]
        top = bottom = padding[1]
    elif len(padding) == 4:
        left = padding[0]
        top = padding[1]
        right = padding[2]
        bottom = padding[3]

    img_np = np.array(img)
    if img_np.shape[-1] == 1:
        fill_value = fill_value[0]
        # BORDER_CONSTANT  BORDER_ISOLATED
    img_np = cv2.copyMakeBorder(img_np, top, bottom, left, right, cv2.BORDER_CONSTANT, value=fill_value)

    return img_np

def erase(np_img, x, y, height, width, erase_value=(128,128,128)):
    """
    Erase a rectangular region in a NumPy image array.

    Args:
        np_img (numpy.ndarray): Input NumPy image array.
        x (int): X-coordinate of the top-left corner of the region to be erased.
        y (int): Y-coordinate of the top-left corner of the region to be erased.
        height (int): Height of the erased region.
        width (int): Width of the erased region.
        erase_value (tuple, optional): The RGB color value to fill the erased region with.
            Default is (128, 128, 128), which corresponds to gray.

    Returns:
        numpy.ndarray: NumPy image array with the specified region erased and filled with the erase_value.
    """
    if not isinstance(np_img, np.ndarray):
        raise TypeError("np_img should be a NumPy array. Got {}.".format(type(np_img)))

    erased_img = np_img.copy()
    erased_img[y:y + height, x:x + width, :] = erase_value
    return erased_img

def augment_Hsv(im, hgain=0.5, sgain=0.5, vgain=0.5):
    # HSV颜色空间增强
    if hgain or sgain or vgain:
        r = np.random.uniform(-1, 1, 3) * [hgain, sgain, vgain] + 1  # random gains
        hue, sat, val = cv2.split(cv2.cvtColor(im, cv2.COLOR_BGR2HSV))
        dtype = im.dtype  # uint8

        x = np.arange(0, 256, dtype=r.dtype)
        lut_hue = ((x * r[0]) % 180).astype(dtype)
        lut_sat = np.clip(x * r[1], 0, 255).astype(dtype)
        lut_val = np.clip(x * r[2], 0, 255).astype(dtype)

        im_hsv = cv2.merge((cv2.LUT(hue, lut_hue), cv2.LUT(sat, lut_sat), cv2.LUT(val, lut_val)))
        return cv2.cvtColor(im_hsv, cv2.COLOR_HSV2BGR, dst=im)

def hist_equalize(im, clahe=True, is_bgr=True):
    # 均衡BGR图像“im”上的直方图，其形状为im(n，m，3)，范围为0-255
    yuv = cv2.cvtColor(im, cv2.COLOR_BGR2YUV if is_bgr else cv2.COLOR_RGB2YUV)
    if clahe:
        c = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        yuv[:, :, 0] = c.apply(yuv[:, :, 0])
    else:
        yuv[:, :, 0] = cv2.equalizeHist(yuv[:, :, 0])
    return cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR if is_bgr else cv2.COLOR_YUV2RGB)


def random_resize_crop(image, target_size, scale_range=(1., 2.)):
    """
    Randomly resize and crop an image.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        target_size (tuple): A tuple (width, height) specifying the target size.
        scale_range (tuple, optional): A tuple (min_scale, max_scale) specifying the range of scaling.
            Default is (0.8, 1.2), which allows resizing between 80% and 120% of the original size.

    Returns:
        Randomly resized and cropped image.
    """
    _check_parameter_is_tuple_2(target_size)
    _check_parameter_is_tuple_2(scale_range)

    min_scale, max_scale = scale_range
    scale_factor = random.uniform(min_scale, max_scale)

    height, width = image.shape[:2]

    new_width = int(width * scale_factor)
    new_height = int(height * scale_factor)

    resized_image = resize(image, (new_width, new_height))

    crop_x = random.randint(0, new_width - target_size[0])
    crop_y = random.randint(0, new_height - target_size[1])
    cropped_image = base_crop2(resized_image, crop_x, crop_y, target_size[0], target_size[1])

    return cropped_image

def random_crop(image, crop_size):
    """
    Randomly crop an image to the specified size.

    Args:
        image (numpy.ndarray): Input image as a NumPy array.
        crop_size (tuple): A tuple (width, height) specifying the crop size.

    Returns:
        Randomly cropped image.
    """
    _check_parameter_is_tuple_2(crop_size)
    height, width = image.shape[:2]
    max_x = width - crop_size[0]
    max_y = height - crop_size[1]

    start_x = random.randint(0, max_x)
    start_y = random.randint(0, max_y)

    cropped_image = base_crop2(image, start_x, start_y, crop_size[0], crop_size[1])

    return cropped_image

def random_horizontal_flip(img, prob=0.5):
    """Randomly flip the input image horizontally."""
    if prob > random.random():
        img = horizontal_flip(img)
    return img

def random_vertical_flip(img, prob=0.5):
    """Randomly flip the input image vertically."""
    if prob > random.random():
        img = vertical_flip(img)
    return img

def random_rotation(img, degrees=None, fill_value=(0, 0, 0)):
    """Randomly rotate the image"""
    if degrees is not None:
        _check_parameter_is_tuple_and_list_2(degrees)
        angle = random.uniform(degrees[0], degrees[1])
    else:
        angle = random.choice([0, 90, 180, 270])

    return rotate(img, angle, fill_value)

def random_lighting(img, alpha):
    """Add AlexNet-style PCA-based noise to an image."""
    alpha_b = np.random.normal(loc=0.0, scale=alpha)
    alpha_g = np.random.normal(loc=0.0, scale=alpha)
    alpha_r = np.random.normal(loc=0.0, scale=alpha)
    table = np.array([
        [55.46 * -0.5675, 4.794 * 0.7192, 1.148 * 0.4009],
        [55.46 * -0.5808, 4.794 * -0.0045, 1.148 * -0.8140],
        [55.46 * -0.5836, 4.794 * -0.6948, 1.148 * 0.4203]
    ])
    pca_b = table[2][0] * alpha_r + table[2][1] * alpha_g + table[2][2] * alpha_b
    pca_g = table[1][0] * alpha_r + table[1][1] * alpha_g + table[1][2] * alpha_b
    pca_r = table[0][0] * alpha_r + table[0][1] * alpha_g + table[0][2] * alpha_b
    img_arr = np.array(img).astype(np.float64)
    img_arr[:, :, 0] += pca_b
    img_arr[:, :, 1] += pca_g
    img_arr[:, :, 2] += pca_r
    img_arr = np.uint8(np.minimum(np.maximum(img_arr, 0), 255))

    return img_arr

def random_apply(img, transforms:list, prob):
    """
    以给定的概率随机应用一个transforms列表
    Args:
        img: Image to be randomly applied a list transformations.
        transforms (list): List of transformations to be applied.
        prob (float): The probability to apply the transformation list.

    Returns:
        Transformed image.
    """
    if prob < random.random():
        return img
    for transform in transforms:
        img = transform(img)
    return img

def random_order(img, transforms:list):
    """
    以随机顺序应用transforms列表.
    Args:
        img: Image to be applied transformations in a random order.
        transforms (list): List of the transformations to be applied.

    Returns:
        Transformed image.
    """
    random.shuffle(transforms)
    for transform in transforms:
        img = transform(img)
    return img

def random_choice(img, transforms:list):
    """
    从transforms列表中随机选择一个变换，并将其应用于图像。
    Args:
        img: Image to be applied transformation.
        transforms (list): List of transformations to be chosen from to apply.

    Returns:
        Transformed image.
    """
    return random.choice(transforms)(img)

def uniform_augment(img, transforms, num_ops):
    """
    Uniformly select and apply a number of transforms sequentially from
    a list of transforms. Randomly assigns a probability to each transform for
    each image to decide whether apply it or not.
    All the transforms in transform list must have the same input/output data type.

    Args:
        img: Image to be applied transformation.
        transforms (list): List of transformations to be chosen from to apply.
        num_ops (int): number of transforms to sequentially aaply.

    Returns:
        Transformed image.

    """
    op_idx = np.random.choice(len(transforms), size=num_ops, replace=False)
    for idx in op_idx:
        augment_op = transforms[idx]
        pr = random.random()
        if random.random() < pr:
            img = augment_op(img.copy())
    return img

class Retinex():
    """
    增强算法Retinex
    - 单尺度 : SSR
    - 多尺度 : MSR
    - 多尺度自适应增益 : MSRCR
    https://blog.csdn.net/m0_62919535/article/details/130372571
    """
    def SSR(self, img, sigma):
        """
        将输入图像转换为了对数空间。/255将像素值归一化到0到1之间，np.log1p取对数并加1是为了避免出现对数运算中分母为0的情况。二维离散傅里叶变换将
        图像从空间域变换到频率域，可以提取出图像中的频率信息。G_recs是用于计算高斯核的半径，result用于最后三通道的叠加。然后循环用于计算加权后的
        频率域图像，再逆二维离散傅里叶变换，得到反射图像，对反射图像进行指数变换，得到最终的输出图像。
        :param img: 输入图像
        :param sigma: 高斯分布的标准差
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        img_fft = np.fft.fft2(img_log)
        G_recs = sigma // 2 + 1
        result = np.zeros_like(img_fft)
        rows, cols, deep = img_fft.shape
        for z in range(deep):
            for i in range(rows):
                for j in range(cols):
                    for k in range(1, G_recs):
                        G = np.exp(-((np.log(k) - np.log(sigma)) ** 2) / (2 * np.log(2) ** 2))
                        #计算高斯滤波器的权值，其中sigma是高斯分布的标准差，k是高斯滤波器的半径，G是高斯滤波器在该点的权值。
                        result[i, j] += G * img_fft[i, j]
        img_ssr = np.real(np.fft.ifft2(result))
        img_ssr = np.exp(img_ssr) - 1
        img_ssr = np.uint8(cv2.normalize(img_ssr, None, 0, 255, cv2.NORM_MINMAX))
        return img_ssr

    def MSR(self, img, scales):
        """
        MSR算法在图像增强中与SSR不同的是，它不需要进行频域变换，它主要是基于图像在多个尺度下的平滑处理和差分处理来提取图像的局部对比度信息和全
        局对比度信息，从而实现对图像的增强。
        在 MSR 算法中，先对图像进行对数变换得到对数图像，然后在不同的尺度下，使用高斯滤波对图像进行平滑处理，得到不同尺度下的平滑图像。接着，通
        过将对数图像和不同尺度下的平滑图像进行差分，得到多个尺度下的细节图像。最后，将这些细节图像加权融合，输出最终的增强图像。

        :param img:
        :param scales: 取值大概在1-10之间
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        result = np.zeros_like(img_log)
        img_light = np.zeros_like(img_log)
        r, c, deep = img_log.shape
        for z in range(deep):
            for scale in scales:
                kernel_size = scale * 4 + 1
                # 高斯滤波器的大小，经验公式kernel_size = scale * 4 + cat
                sigma = scale
                img_smooth = cv2.GaussianBlur(img_log[:, :, z], (kernel_size, kernel_size), sigma)
                img_detail = img_log[:, :, z] - img_smooth
                result[:, :, z] += cv2.resize(img_detail, (c, r))
                img_light[:, :, z] += cv2.resize(img_smooth, (c, r))
        img_msr = np.exp(result+img_light) - 1
        img_msr = np.uint8(cv2.normalize(img_msr, None, 0, 255, cv2.NORM_MINMAX))
        return img_msr


    def MSRCR(self, img, scales, k):
        """
        使用 MSRCR 算法的步骤：
            1. 将输入图像转换为对数空间（log-space）。
            2. 对图像在不同尺度下进行平滑处理，得到不同尺度下的平滑图像。
            3. 计算细节图像，通过将对数图像与平滑图像相减获得。
            4. 对细节图像进行缩放，以匹配输入图像的大小。
            5. 根据尺度因子调整细节图像的权重。
            6. 将权重调整后的细节图像与平滑图像相加，得到增强的图像。
            7. 最后，将增强的图像反转对数变换，以还原原始图像的像素值范围。
        :param img:
        :param scales:取值大概在1-10之间
        :param k: k的取值范围在10~20之间比较合适。当k取值较小时，图像的细节增强效果比较明显，但会出现较强的噪点，当k取值较大时，图像的细节
                    增强效果不明显，但噪点会减少。
        :return:
        """
        img_log = np.log1p(np.array(img, dtype="float") / 255)
        result = np.zeros_like(img_log)
        img_light = np.zeros_like(img_log)
        r, c, deep = img_log.shape
        for z in range(deep):
            for scale in scales:
                kernel_size = scale * 4 + 1
                # 高斯滤波器的大小，经验公式kernel_size = scale * 4 + cat
                sigma = scale
                G_ratio=sigma**2/(sigma**2+k)
                img_smooth = cv2.GaussianBlur(img_log[:, :, z], (kernel_size, kernel_size), sigma)
                img_detail = img_log[:, :, z] - img_smooth
                result[:, :, z] += cv2.resize(img_detail, (c, r))
                result[:, :, z]=result[:, :, z]*G_ratio
                img_light[:, :, z] += cv2.resize(img_smooth, (c, r))

        img_msrcr = np.exp(result+img_light) - 1
        img_msrcr = np.uint8(cv2.normalize(img_msrcr, None, 0, 255, cv2.NORM_MINMAX))
        return img_msrcr

def blur(img, ksize):
    """均值滤波 """
    blur_img = cv2.blur(img, ksize=ksize)
    return blur_img

def median_blur(img, ksize):
    """中值滤波"""
    if img.dtype == np.float32 and ksize not in {3, 5, 7}:
        raise ValueError(f"Invalid ksize value {ksize}.The available values are 3, 5, and 7")
    medblur_img = cv2.medianBlur(img, ksize=ksize)
    return medblur_img

def gaussian_blur(img, ksize):
    """
    高斯模糊,提供给不熟悉高斯模糊参数的用户,sigma根据ksize进行自动计算,具体可以参考下面
    https://docs.opencv.org/master/d4/d13/tutorial_py_filtering.html
    """
    sigma = 0.3*((ksize-1)*0.5 - 1) + 0.8
    gaussianblur_img = cv2.GaussianBlur(img, ksize=(ksize,ksize), sigmaX=sigma)
    return gaussianblur_img

def bilateral_blur(img, d=5, sigma_color=75, sigma_space=75):
    """双边滤波"""
    return cv2.bilateralFilter(img, d, sigma_color, sigma_space)

class Filter():
    """手写实现,仅供学习参考,最好还是使用cv2
    https://blog.csdn.net/m0_62919535/category_11936595.html?spm=1001.2014.3001.5482
    """
    def median_filtering(self, img,ksize=3):
        """
        中值滤波
        :param img:输入图像
        :param ksize: 核大小
        :return: 中值滤波平滑
        """
        h, w, c = img.shape
        half = ksize//2
        dst = np.zeros((h+2*half,w+2*half,c),np.uint8)
        dst[half:half+h, half:half+w] = img.copy()

        tmp=dst.copy()
        for y in range(h):
            for x in range(w):
                for z in range(c):
                    dst[half+x,half+y]=np.median(tmp[x:x+ksize,y:y+ksize])
        output=dst[half:half+h,half:half+w]
        return output

    def Arerage_Filtering(self, img, k_size=3):
        """
        均值滤波函数,默认会返回灰度图，因为三个for循环实在是太耗费时间了。而且，在这里需要考虑到边界点的问题。
        计算填充的宽度，即卷积核宽度的一半，用于处理图像边缘。使用cv2.copyMakeBorder函数进行边缘填充，将图
        像的边缘复制并填充到周围，以防止边缘像素点无法进行卷积。
        :param img: 原始图像，要求为灰度图
        :param k_size: 滤波核大小,默认为3,确保滤波核大小为奇数
        :return: 处理后的均值滤波图像
        """
        if k_size % 2 == 0:
            k_size += 1
        rows, cols = img.shape[:2]
        # 计算需要在图像边界扩充的大小
        pad_width = (k_size - 1) // 2
        # 在图像边界进行扩充
        img_pad = cv2.copyMakeBorder(img, pad_width, pad_width, pad_width, pad_width, cv2.BORDER_REPLICATE)
        img_filter = np.zeros_like(img)
        for i in range(rows):
            for j in range(cols):
                pixel_values = img_pad[i:i+k_size, j:j+k_size].flatten()
                img_filter[i, j] = np.mean(pixel_values)

        return img_filter

    def gaussian_kernel(self, size, sigma):
        """
        生成高斯核
        :param size: 核的大小
        :param sigma: 标准差
        :return:
        """
        kernel = np.zeros((size, size), dtype=np.float32)
        center = size // 2
        for i in range(size):
            for j in range(size):
                x, y = i - center, j - center
                kernel[i, j] = np.exp(-(x**2 + y**2)/(2*sigma**2))
        kernel /= 2 * np.pi * sigma**2
        kernel /= np.sum(kernel)
        return kernel

    def Gaussian_Filtering(self, img, kernel_size, sigma):
        """
        生成高斯滤波
        :param img: 输入图像
        :param kernel_size: 核大小
        :param sigma: 标准差
        :return:
        """
        kernel = self.gaussian_kernel(kernel_size, sigma)
        height, width, _ = img.shape
        result = np.zeros_like(img, dtype=np.float32)

        pad_size = kernel_size // 2
        img_pad = np.pad(img, [(pad_size, pad_size), (pad_size, pad_size), (0, 0)], mode='constant')
        for c in range(_):
            for i in range(pad_size, height + pad_size):
                for j in range(pad_size, width + pad_size):
                    result[i - pad_size, j - pad_size, c] = np.sum(kernel * img_pad[i - pad_size:i + pad_size + 1, j - pad_size:j + pad_size + 1, c])
        return np.uint8(result)


"""
Copyright (c) 2024, Auorui.
All rights reserved.

Supports one hot encoding for np and torch.
https://blog.csdn.net/m0_62919535/article/details/135516163?spm=1001.2014.3001.5501

time 2024-01-11
"""
import torch
import numpy as np
from pyzjr.core.general import is_tensor, is_numpy

__all__ = ["one_hot", "get_one_hot", "get_one_hot_with_torch", "get_one_hot_with_np"]

def one_hot(labels, num_classes=-1):
    """
    将标签转为独热编码, 经过测试与torch.nn.functional里面的函数测试相同
    :param labels: 标签
    :param num_classes: 默认为-1, 表示进行自动计算类别最大的那个
    Examples:
        >>> label_1 = torch.arange(0, 5) % 3
        tensor([0, 1, 2, 0, 1])
        >>> label_2 = torch.arange(0, 6).view(3, 2) % 3
        tensor([[0, 1], [2, 0], [1, 2]])
        >>> print(one_hot(label_1))
        tensor([[1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0]])
        >>> print(one_hot(label_1, 5))
        tensor([[1, 0, 0, 0, 0],
                [0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0],
                [1, 0, 0, 0, 0],
                [0, 1, 0, 0, 0]])
        >>> print(one_hot(label_2))
        tensor([[[1, 0, 0],
                 [0, 1, 0]],
                [[0, 0, 1],
                 [1, 0, 0]],
                [[0, 1, 0],
                 [0, 0, 1]]])
    """
    if num_classes == -1:
        num_classes = int(labels.max()) + 1
    one_hot_tensor = torch.zeros(labels.size() + (num_classes,), dtype=torch.int64)
    one_hot_tensor.scatter_(-1, labels.unsqueeze(-1).to(torch.int64), 1)
    return one_hot_tensor

def get_one_hot_with_torch(labels, num_classes=-1):
    """用于分割网络的one hot"""
    labels = torch.as_tensor(labels)
    ones = one_hot(labels, num_classes)
    return ones.view(*labels.size(), num_classes)

def get_one_hot_with_np(labels, num_classes=-1):
    """部分时候用于处理标签还没有转为tensor的情况"""
    if num_classes == -1:
        num_classes = int(np.max(labels)) + 1
    input_shape = labels.shape
    seg_labels  = np.eye(num_classes)[labels.reshape([-1])]
    seg_labels  = seg_labels.reshape((int(input_shape[0]), int(input_shape[1]), num_classes))
    return seg_labels

def get_one_hot(labels, num_classes=-1):
    if is_numpy(labels):
        return get_one_hot_with_np(labels=labels, num_classes=num_classes)
    if is_tensor(labels):
        return get_one_hot_with_torch(labels=labels, num_classes=num_classes)

if __name__=="__main__":
    seg_labels = np.random.randint(0, 3, size=[512, 512])
    seg_labels2 = torch.randint(0, 3, size=[512, 512])
    print(get_one_hot(seg_labels))
    print(get_one_hot(seg_labels).shape)
    print(get_one_hot(seg_labels2))   # torch.Size([512, 512, 3])
    print(get_one_hot(seg_labels2).shape)


import os
import torch
import gc

def cpu():
    """
    Returns the CPU device.

    Returns:
        torch.device: The CPU device.
    """
    return torch.device('cpu')

def gpu(i=0, cuda=True):
    """
    If cuda is true and gpu is available, return gpu (i); otherwise, return cpu()
    :param i: Index i, indicating which GPU block to use
    :param cuda: Boolean value indicating whether to use cuda, if not, CPU can be used, set to False
    """
    if cuda and torch.cuda.is_available() and torch.cuda.device_count() >= i + 1:
        return torch.device(f'cuda:{i}')
    return torch.device('cpu')

def allgpu():
    """
    Those who can use this function must be very rich.
    """
    devices = []
    for i in range(torch.cuda.device_count()):
        devices.append(i)
    return devices if devices else [torch.device('cpu')]

def num_workers(batch_size):
    """
    确定用于数据加载的并行工作进程的数量
    """
    num_workers = min([os.cpu_count(), batch_size if batch_size > 1 else 0, 8])
    return num_workers

def release_memory(*args):
    """
    Function to release memory resources, particularly useful when working with PyTorch and CUDA.
    """
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    if not isinstance(args, list):
        args = list(args)
    for i in range(len(args)):
        args[i] = None
    gc.collect()
    return args





from .general import (
    is_pil,
    is_numpy,
    is_gray_image,
    is_rgb_image,
    get_num_channels,
    get_image_size,
)

from ._utils import (
    Module,
    ModuleList,
    Parameter,
    Tensor,
    arange,
    as_tensor,
    complex,
    concatenate,
    diag,
    einsum,
    eye,
    linspace,
    normalize,
    ones,
    ones_like,
    pad,
    rand,
    softmax,
    stack,
    tensor,
    where,
    zeros,
    zeros_like,
)

from .error import *
from .lr_scheduler import _LRScheduler

__all__ = [
    "arange",
    "concatenate",
    "Module",
    "ModuleList",
    "Tensor",
    "tensor",
    "Parameter",
    "normalize",
    "pad",
    "stack",
    "softmax",
    "as_tensor",
    "rand",
    "where",
    "eye",
    "ones",
    "ones_like",
    "einsum",
    "zeros",
    "complex",
    "zeros_like",
    "linspace",
    "diag",
]




"""
Copyright (c) 2024, Auorui.
All rights reserved.

v1: <https://arxiv.org/pdf/1704.04861.pdf>
v2: <https://arxiv.org/pdf/1801.04381.pdf>
v3: <https://arxiv.org/abs/1905.02244.pdf>
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial

__all__ = ["MobileNetV1", "MobileNetV2", "MobileNetV3", "MobileNetV3_Small", "MobileNetV3_Large"]

class DepthSepConv(nn.Module):
    """
    深度可分卷积: DW卷积 + PW卷积
    dw卷积, 当分组个数等于输入通道数时, 输出矩阵的通道输也变成了输入通道数
    pw卷积, 使用了1x1的卷积核与普通的卷积一样
    """
    def __init__(self, in_channels, out_channels, stride):
        super(DepthSepConv, self).__init__()
        self.depthwise = nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=stride, groups=in_channels, padding=1)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.batch_norm1 = nn.BatchNorm2d(in_channels)
        self.batch_norm2 = nn.BatchNorm2d(out_channels)
        self.relu6 = nn.ReLU6(inplace=True)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.batch_norm1(x)
        x = self.relu6(x)
        x = self.pointwise(x)
        x = self.batch_norm2(x)
        x = self.relu6(x)

        return x

class MobileNetV1(nn.Module):
    def __init__(self, num_classes=1000, drop_rate=0.2):
        super(MobileNetV1, self).__init__()
        # torch.Size([1, 3, 224, 224])
        self.conv_bn = nn.Sequential(
                nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, stride=2, padding=1, bias=False),
                nn.BatchNorm2d(32),
                nn.ReLU(inplace=True)
            )                                # torch.Size([1, 32, 112, 112])
        self.dwmodule = nn.Sequential(
            # 参考MobileNet_V1 https://arxiv.org/pdf/1704.04861.pdf Table 1
            DepthSepConv(32, 64, 1),            # torch.Size([1, 64, 112, 112])
            DepthSepConv(64, 128, 2),           # torch.Size([1, 128, 56, 56])
            DepthSepConv(128, 128, 1),          # torch.Size([1, 128, 56, 56])
            DepthSepConv(128, 256, 2),          # torch.Size([1, 256, 28, 28])
            DepthSepConv(256, 256, 1),          # torch.Size([1, 256, 28, 28])
            DepthSepConv(256, 512, 2),          # torch.Size([1, 512, 14, 14])
            # 5 x DepthSepConv(512, 512, 1),
            DepthSepConv(512, 512, 1),          # torch.Size([1, 512, 14, 14])
            DepthSepConv(512, 512, 1),
            DepthSepConv(512, 512, 1),
            DepthSepConv(512, 512, 1),
            DepthSepConv(512, 512, 1),
            DepthSepConv(512, 1024, 2),         # torch.Size([1, 1024, 7, 7])
            DepthSepConv(1024, 1024, 1),
            nn.AvgPool2d(7, stride=1),
        )
        self.fc = nn.Linear(in_features=1024, out_features=num_classes)
        self.dropout = nn.Dropout(p=drop_rate)
        self.softmax = nn.Softmax(dim=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.conv_bn(x)
        x = self.dwmodule(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        x = self.softmax(self.dropout(x))
        return x

def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.It can be seen here:
        https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    Args:
        v: The number of input channels.
        divisor: The number of channels should be a multiple of this value.
        min_value: The minimum value of the number of channels, which defaults to the advisor.

    Returns: It ensures that all layers have a channel number that is divisible by 8
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

class ConvBNActivation(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, groups=1,
                 norm_layer=None, activation_layer=None, dilation=1,):
        super(ConvBNActivation, self).__init__()
        padding = (kernel_size - 1) // 2 * dilation
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if activation_layer is None:
            activation_layer = nn.ReLU6
        self.convbnact=nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, dilation=dilation, groups=groups,
                      bias=False),
            norm_layer(out_planes),
            activation_layer(inplace=True)
        )
        self.out_channels = out_planes

    def forward(self, x):
        return self.convbnact(x)

class InvertedResidualv2(nn.Module):
    def __init__(self, in_planes, out_planes, stride, expand_ratio):
        super(InvertedResidualv2, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(in_planes * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_planes == out_planes

        layers = []
        if expand_ratio != 1:
            # pw 利用1x1卷积进行通道数的上升
            layers.append(ConvBNActivation(in_planes, hidden_dim, kernel_size=1))

        layers.extend([
            # dw 进行3x3的逐层卷积，进行跨特征点的特征提取
            ConvBNActivation(hidden_dim, hidden_dim, kernel_size=3, stride=stride, groups=hidden_dim),
            # pw-linear 利用1x1卷积进行通道数的下降
            nn.Conv2d(hidden_dim, out_planes, kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(out_planes),
        ])
        self.conv = nn.Sequential(*layers)
        self.out_channels = out_planes

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(self, num_classes=1000, drop_rate=0.2, width_mult=1.0, round_nearest=8):
        """
        MobileNet V2 main class

        Args:
            num_classes (int): Number of classes
            drop_rate (float): Dropout layer drop rate
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
        """
        super(MobileNetV2, self).__init__()
        input_channel = 32
        last_channel = 1280
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        last_channel = _make_divisible(last_channel * max(1.0, width_mult), round_nearest)
        inverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]
        # t表示是否进行1*1卷积上升的过程 c表示output_channel大小 n表示小列表倒残差次数 s是步长,表示是否对高和宽进行压缩
        # building first layer
        features = [ConvBNActivation(3, input_channel, stride=2)]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(InvertedResidualv2(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        features.append(ConvBNActivation(input_channel, last_channel, kernel_size=1))
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        self.classifier = nn.Sequential(
            nn.Dropout(drop_rate),
            nn.Linear(last_channel, num_classes),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.features(x)
        # Cannot use "squeeze" as batch-size can be 1 => must use reshape with x.shape[0]
        x = F.adaptive_avg_pool2d(x, (1, 1)).reshape(x.shape[0], -1)
        x = self.classifier(x)
        return x

class SeModule(nn.Module):
    def __init__(self, input_channels, reduction=4):
        super(SeModule, self).__init__()
        expand_size = _make_divisible(input_channels // reduction, 8)
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(input_channels, expand_size, kernel_size=1, bias=False),
            nn.BatchNorm2d(expand_size),
            nn.ReLU(inplace=True),
            nn.Conv2d(expand_size, input_channels, kernel_size=1, bias=False),
            nn.Hardsigmoid()
        )

    def forward(self, x):
        return x * self.se(x)

class MobileNetV3(nn.Module):
    """
    MobileNet V3 main class

    Args:
        num_classes: Number of classes
        mode: "large" or "small"
    """
    def __init__(self, num_classes=1000, mode=None, drop_rate=0.2):
        super().__init__()
        norm_layer = partial(nn.BatchNorm2d, eps=0.001, momentum=0.01)
        layers = []
        inverted_residual_setting, last_channel = _mobilenetv3_cfg[mode]
        # building first layer
        firstconv_output_channels = 16
        layers.append(ConvBNActivation(3, firstconv_output_channels, kernel_size=3, stride=2, norm_layer=norm_layer,
                                       activation_layer=nn.Hardswish))
        layers.append(inverted_residual_setting)
        # building last several layers
        lastconv_input_channels = 96 if mode == "small" else 160
        lastconv_output_channels = 6 * lastconv_input_channels
        layers.append(ConvBNActivation(lastconv_input_channels, lastconv_output_channels, kernel_size=1,
                                       norm_layer=norm_layer, activation_layer=nn.Hardswish))

        self.features = nn.Sequential(*layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Sequential(
            nn.Linear(lastconv_output_channels, last_channel),
            nn.Hardswish(inplace=True),
            nn.Dropout(p=drop_rate, inplace=True),
            nn.Linear(last_channel, num_classes),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.features(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)

        return x

class InvertedResidualv3(nn.Module):
    '''expand + depthwise + pointwise'''
    def __init__(self, kernel_size, input_channels, expanded_channels, out_channels, activation, use_se, stride):
        super(InvertedResidualv3, self).__init__()
        self.stride = stride
        norm_layer = partial(nn.BatchNorm2d, eps=0.001, momentum=0.01)
        self.use_res_connect = stride == 1 and input_channels == out_channels
        activation_layer = nn.ReLU if activation == "RE" else nn.Hardswish
        layers = []
        if expanded_channels != input_channels:
            layers.append(ConvBNActivation(input_channels, expanded_channels, kernel_size=1,
                                           norm_layer=norm_layer, activation_layer=activation_layer))

        # depthwise
        layers.append(ConvBNActivation(expanded_channels, expanded_channels, kernel_size=kernel_size,
                                       stride=stride, groups=expanded_channels,
                                       norm_layer=norm_layer, activation_layer=activation_layer))
        if use_se:
            layers.append(SeModule(expanded_channels))

        layers.append(ConvBNActivation(expanded_channels, out_channels, kernel_size=1, norm_layer=norm_layer,
                                       activation_layer=nn.Identity))

        self.block = nn.Sequential(*layers)
        self.out_channels = out_channels

    def forward(self, x):
        result = self.block(x)
        if self.use_res_connect:
            result += x
        return result

_mobilenetv3_cfg = {
    "large": [nn.Sequential(
        # kernel, in_chs, exp_chs, out_chs, act, use_se, stride
        InvertedResidualv3(3, 16, 16, 16, "RE", False, 1),
        InvertedResidualv3(3, 16, 64, 24, "RE", False, 2),
        InvertedResidualv3(3, 24, 72, 24, "RE", False, 1),
        InvertedResidualv3(5, 24, 72, 40, "RE", True, 2),
        InvertedResidualv3(5, 40, 120, 40, "RE", True, 1),
        InvertedResidualv3(5, 40, 120, 40, "RE", True, 1),
        InvertedResidualv3(3, 40, 240, 80, "HS", False, 2),
        InvertedResidualv3(3, 80, 200, 80, "HS", False, 1),
        InvertedResidualv3(3, 80, 184, 80, "HS", False, 1),
        InvertedResidualv3(3, 80, 184, 80, "HS", False, 1),
        InvertedResidualv3(3, 80, 480, 112, "HS", True, 1),
        InvertedResidualv3(3, 112, 672, 112, "HS", True, 1),
        InvertedResidualv3(5, 112, 672, 160, "HS", True, 1),
        InvertedResidualv3(5, 160, 672, 160, "HS", True, 2),
        InvertedResidualv3(5, 160, 960, 160, "HS", True, 1),
    ),
        _make_divisible(1280, 8)],
    "small": [nn.Sequential(
        # kernel, in_chs, exp_chs, out_chs, act, use_se, stride
        InvertedResidualv3(3, 16, 16, 16, "RE", True, 2),
        InvertedResidualv3(3, 16, 72, 24, "RE", False, 2),
        InvertedResidualv3(3, 24, 88, 24, "RE", False, 1),
        InvertedResidualv3(5, 24, 96, 40, "HS", True, 2),
        InvertedResidualv3(5, 40, 240, 40, "HS", True, 1),
        InvertedResidualv3(5, 40, 240, 40, "HS", True, 1),
        InvertedResidualv3(5, 40, 120, 48, "HS", True, 1),
        InvertedResidualv3(5, 48, 144, 48, "HS", True, 1),
        InvertedResidualv3(5, 48, 288, 96, "HS", True, 2),
        InvertedResidualv3(5, 96, 576, 96, "HS", True, 1),
        InvertedResidualv3(5, 96, 576, 96, "HS", True, 1),
    ),
        _make_divisible(1024, 8)],
}

def MobileNetV3_Large(num_classes):
    """Large version of mobilenet_v3"""
    return MobileNetV3(num_classes=num_classes, mode="large")

def MobileNetV3_Small(num_classes):
    """small version of mobilenet_v3"""
    return MobileNetV3(num_classes=num_classes, mode="small")

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = MobileNetV2(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # v1 Total params: 3,221,988
    # v2 Total params: 2,230,500
    # v3_small Total params: 1,520,252
    # v3_large Total params: 3,868,460


"""
Original paper address https://arxiv.org/pdf/1709.01507.pdf
Squeeze-and-Excitation Networks
Obtained from https://github.com/moskomule/senet.pytorch/blob/master/senet
"""
import torch.nn as nn
from pyzjr.Models.backbone.resnet import ResNet

__all__ = ["se_resnet18", "se_resnet34", "se_resnet50", "se_resnet101", "se_resnet152"]

class SELayer(nn.Module):
    def __init__(self, channel, reduction=16):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)

def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)

def conv1x1(in_channels, out_channels, stride=1):
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

class SEBasicBlock(nn.Module):
    expansion = 1
    def __init__(self, inplanes, planes, stride=1, downsample=False, reduction=16):
        super(SEBasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, 1)
        self.bn2 = nn.BatchNorm2d(planes)
        self.se = SELayer(planes, reduction)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(inplanes, planes * self.expansion, self.stride),
            nn.BatchNorm2d(planes * self.expansion),
        )

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.se(out)

        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out


class SEBottleneck(nn.Module):
    expansion = 4
    def __init__(self, inplanes, planes, stride=1, downsample=False, reduction=16):
        super(SEBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.se = SELayer(planes * 4, reduction)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(inplanes, planes * self.expansion, self.stride),
            nn.BatchNorm2d(planes * self.expansion),
        )

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        out = self.se(out)

        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out


def se_resnet18(num_classes):
    model = ResNet(SEBasicBlock, [2, 2, 2, 2], num_classes=num_classes)
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    return model


def se_resnet34(num_classes):
    model = ResNet(SEBasicBlock, [3, 4, 6, 3], num_classes=num_classes)
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    return model


def se_resnet50(num_classes):
    model = ResNet(SEBottleneck, [3, 4, 6, 3], num_classes=num_classes)
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    return model


def se_resnet101(num_classes):
    model = ResNet(SEBottleneck, [3, 4, 23, 3], num_classes=num_classes)
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    return model


def se_resnet152(num_classes):
    model = ResNet(SEBottleneck, [3, 8, 36, 3], num_classes=num_classes)
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    return model

if __name__=="__main__":
    import torchsummary
    import torch
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = se_resnet50(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 26,031,172


"""
The source code of torch, from torch.optim.lr_Scheduler import _LRScheduler
But it cannot be directly called because its __all__ Not inside_ LRScheduler
I copied it directly from the source code here.
"""
from functools import wraps
import warnings
import weakref
from torch.optim.optimizer import Optimizer

__all__ = ["_LRScheduler"]

EPOCH_DEPRECATION_WARNING = (
    "The epoch parameter in `scheduler.step()` was not necessary and is being "
    "deprecated where possible. Please use `scheduler.step()` to step the "
    "scheduler. During the deprecation, if epoch is different from None, the "
    "closed form is used instead of the new chainable form, where available. "
    "Please open an issue if you are unable to replicate your use case: "
    "https://github.com/pytorch/pytorch/issues/new/choose."
)

class _LRScheduler(object):

    def __init__(self, optimizer, last_epoch=-1, verbose=False):

        # Attach optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        # Initialize epoch and base learning rates
        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified "
                                   "in param_groups[{}] when resuming an optimizer".format(i))
        self.base_lrs = [group['initial_lr'] for group in optimizer.param_groups]
        self.last_epoch = last_epoch

        # Following https://github.com/pytorch/pytorch/issues/20124
        # We would like to ensure that `lr_scheduler.step()` is called after
        # `optimizer.step()`
        def with_counter(method):
            if getattr(method, '_with_counter', False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # Keep a weak reference to the optimizer instance to prevent
            # cyclic references.
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned function here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper

        self.optimizer.step = with_counter(self.optimizer.step)
        self.verbose = verbose

        self._initial_step()

    def _initial_step(self):
        """Initialize step counts and performs a step"""
        self.optimizer._step_count = 0
        self._step_count = 0
        self.step()

    def state_dict(self):
        """Returns the state of the scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

    def load_state_dict(self, state_dict):
        """Loads the schedulers state.

        Args:
            state_dict (dict): scheduler state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)

    def get_last_lr(self):
        """ Return last computed learning rate by current scheduler.
        """
        return self._last_lr

    def get_lr(self):
        # Compute learning rate using chainable form of the scheduler
        raise NotImplementedError

    def print_lr(self, is_verbose, group, lr, epoch=None):
        """Display the current learning rate.
        """
        if is_verbose:
            if epoch is None:
                print('Adjusting learning rate'
                      ' of group {} to {:.4e}.'.format(group, lr))
            else:
                epoch_str = ("%.2f" if isinstance(epoch, float) else
                             "%.5d") % epoch
                print('Epoch {}: adjusting learning rate'
                      ' of group {} to {:.4e}.'.format(epoch_str, group, lr))


    def step(self, epoch=None):
        # Raise a warning if old pattern is detected
        # https://github.com/pytorch/pytorch/issues/20124
        if self._step_count == 1:
            if not hasattr(self.optimizer.step, "_with_counter"):
                warnings.warn("Seems like `optimizer.step()` has been overridden after learning rate scheduler "
                              "initialization. Please, make sure to call `optimizer.step()` before "
                              "`lr_scheduler.step()`. See more details at "
                              "https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate", UserWarning)

            # Just check if there were two first lr_scheduler.step() calls before optimizer.step()
            elif self.optimizer._step_count < 1:
                warnings.warn("Detected call of `lr_scheduler.step()` before `optimizer.step()`. "
                              "In PyTorch 1.1.0 and later, you should call them in the opposite order: "
                              "`optimizer.step()` before `lr_scheduler.step()`.  Failure to do this "
                              "will result in PyTorch skipping the first value of the learning rate schedule. "
                              "See more details at "
                              "https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate", UserWarning)
        self._step_count += 1

        class _enable_get_lr_call:

            def __init__(self, o):
                self.o = o

            def __enter__(self):
                self.o._get_lr_called_within_step = True
                return self

            def __exit__(self, type, value, traceback):
                self.o._get_lr_called_within_step = False

        with _enable_get_lr_call(self):
            if epoch is None:
                self.last_epoch += 1
                values = self.get_lr()
            else:
                warnings.warn(EPOCH_DEPRECATION_WARNING, UserWarning)
                self.last_epoch = epoch
                if hasattr(self, "_get_closed_form_lr"):
                    values = self._get_closed_form_lr()
                else:
                    values = self.get_lr()

        for i, data in enumerate(zip(self.optimizer.param_groups, values)):
            param_group, lr = data
            param_group['lr'] = lr
            self.print_lr(self.verbose, i, lr, epoch)

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]


from .Standardpooling import MaxPool2d, AvgPool2d
from .HaarWavelet import HWDownsampling
from .StrideConv import StridedConvolution
from .strippooling import StripPooling



__all__ = ["MaxPool2d", "AvgPool2d", "HWDownsampling", "StridedConvolution", "StripPooling"]


"""
Copyright (c) 2024, Auorui.
All rights reserved.
用于 语义分割 的指标检测
                 | Predicted Positive | Predicted Negative |
Actual Positive  |        TP          |         FN         |
Actual Negative  |        FP          |         TN         |
Time: 2024-01-26
"""
import torch
import numpy as np

__all__ = ["Miou", "Recall", "Precision", "F1Score", "DiceCoefficient", "Accuracy",
           "SegmentationIndex", "AIU"]

class Miou(object):
    """
    iou = ( A & B ) / ( A | B )
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]

        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        pred = pred > self.threshold
        true = true > self.threshold
        intersection = (pred & true).sum(axis=(1,2,3))
        union = (pred | true).sum(axis=(1,2,3))

        iou = intersection / (union + self.esp)
        return iou.mean()

class Recall(object):
    """
    recall = TP / (TP + FN)
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]

        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        pred = pred > self.threshold
        true = true > self.threshold

        TP = (pred & true).sum(axis=(1, 2, 3))
        FN = (~pred & true).sum(axis=(1, 2, 3))

        recall = TP / (TP + FN + self.esp)
        return recall.mean()

class Precision(object):
    """
    precision = TP / (TP + FP)
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]

        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        pred = pred > self.threshold
        true = true > self.threshold

        TP = (pred & true).sum(axis=(1, 2, 3))
        FP = (pred & ~true).sum(axis=(1, 2, 3))

        precision = TP / (TP + FP + self.esp)
        return precision.mean()

class Accuracy(object):
    """
    accuracy = (TP + TN) / (TP + TN + FP + FN)
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]
        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        pred = pred > self.threshold
        true = true > self.threshold

        TP = (pred & true).sum(axis=(1, 2, 3))
        TN = (~pred & ~true).sum(axis=(1, 2, 3))
        FP = (pred & ~true).sum(axis=(1, 2, 3))
        FN = (~pred & true).sum(axis=(1, 2, 3))

        accuracy = (TP + TN) / (TP + TN + FP + FN + self.esp)
        return accuracy.mean()

class F1Score(object):
    """
    f1score = 2 * (precision * recall) / (precision + recall)
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]
        pred = pred > self.threshold
        true = true > self.threshold

        recall = Recall(self.num_classes)(pred, true)
        precision = Precision(self.num_classes)(pred, true)
        f1_score = 2 * (precision * recall) / (precision + recall + self.esp)
        return f1_score.mean()

class DiceCoefficient(object):
    """
    dice_coefficient = (2 * TP) / (2 * TP + FP + FN)
    """
    def __init__(self, num_classes, ignore=None, threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().numpy()
        if torch.is_tensor(true):
            true = true.cpu().numpy()
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]
        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        pred = pred > self.threshold
        true = true > self.threshold

        TP = (pred & true).sum(axis=(1, 2, 3))
        FP = (pred & ~true).sum(axis=(1, 2, 3))
        FN = (~pred & true).sum(axis=(1, 2, 3))

        dice_coefficient = (2 * TP) / (2 * TP + FP + FN + self.esp)
        return dice_coefficient.mean()

class SegmentationIndex(object):
    def __init__(self, num_classes, ignore=None, headline="Semantic Segmentation Index:",threshold=.5, esp=1e-5):
        self.num_classes = num_classes
        self.threshold = threshold
        self.ignore = ignore
        self.esp = esp
        self.miou = Miou(self.num_classes, self.ignore)
        self.recall = Recall(self.num_classes, self.ignore)
        self.precision = Precision(self.num_classes, self.ignore)
        self.f1Score = F1Score(self.num_classes, self.ignore)
        self.dice_coefficient = DiceCoefficient(self.num_classes, self.ignore)
        self.acc = Accuracy(self.num_classes, self.ignore)
        self.index = []
        self.headline = headline

    def update(self, pred, true):
        pred = pred.cpu().numpy() if torch.is_tensor(pred) else pred
        true = true.cpu().numpy() if torch.is_tensor(true) else true
        miou_value = self.miou(pred, true)
        recall_value = self.recall(pred, true)
        precision_value = self.precision(pred, true)
        f1score_value = self.f1Score(pred, true)
        dice_value = self.dice_coefficient(pred, true)
        acc_value = self.acc(pred, true)

        self.index.append({
            'miou': miou_value,
            'recall': recall_value,
            'precision': precision_value,
            'f1score': f1score_value,
            'dice': dice_value,
            'accuracy': acc_value
        })
        return miou_value, recall_value, precision_value, f1score_value, dice_value, acc_value

    def __str__(self):
        latest_index = self.index[-1] if self.index else None

        return (f"\033[94m{self.headline}\n"
                f"\t  Miou         Recall         Precision         F1Score         "
                f"DiceCoefficient         Accuracy\n\033[94m"
                f"\t{latest_index['miou']:.5f}       {latest_index['recall']:.5f}          "
                f"{latest_index['precision']:.5f}          {latest_index['f1score']:.5f}            "
                f"{latest_index['dice']:.5f}              {latest_index['accuracy']:.5f}\033[0m")


class AIU(object):
    """
    Reference from paper https://arxiv.org/pdf/1901.06340.pdf Formula 7:
    aiu = [Nt_pg / (Nt_p + Nt_g - Nt_pg] / N
    """
    def __init__(self, num_classes, ignore=None, threshold=(1, 100), esp=1e-5):
        self.num_classes = num_classes
        self.thresholds = [i / 100.0 for i in range(*threshold)]  # 从0.01到0.99，间隔0.01
        self.esp = esp
        self.ignore = ignore

    def __call__(self, pred, true):
        total_aiu = 0.
        pred = pred.cpu().numpy() if torch.is_tensor(pred) else pred
        true = true.cpu().numpy() if torch.is_tensor(true) else true
        if self.ignore:
            pred = pred[:, :self.ignore, :, :]
            true = true[:, :self.ignore, :, :]

        assert pred.shape[1] == true.shape[1] == self.num_classes, "Number of classes mismatch"
        for threshold in self.thresholds:
            intersection = (pred > threshold) & (true > threshold)
            Nt_pg = intersection.sum(axis=(1, 2, 3))
            Nt_p = (pred > threshold).sum(axis=(1, 2, 3))
            Nt_g = (true > threshold).sum(axis=(1, 2, 3))

            aiu_t = Nt_pg / (Nt_p + Nt_g - Nt_pg + self.esp)  # 添加小的常数以避免除零错误
            total_aiu += aiu_t
        average_aiu = total_aiu / len(self.thresholds)
        return np.array(average_aiu).mean()

if __name__ == "__main__":
    batch_size, num_classes, height, width = 2, 2, 3, 3

    # pred = torch.randn((batch_size, num_classes, height, width))
    # true = torch.randint(0, 2, (batch_size, num_classes, height, width))
    # print(pred, true)
    pred = torch.Tensor([[[[-0.1670,  0.1940,  0.4754],
                           [-0.0729,  1.1101, -0.4133],
                           [ 0.1728,  0.2493, -0.8165]],

                          [[ 0.7978,  0.7136, -0.9680],
                           [ 0.1166,  0.2708,  1.5798],
                           [ 0.5725,  0.5781, -0.3959]]]])
    print(pred.shape)
    true = torch.Tensor([[[[0, 0, 0],
                           [0, 0, 1],
                           [1, 1, 0]],

                          [[1, 1, 1],
                           [0, 0, 1],
                           [1, 1, 0]],

                          # [[0, 0, 0],
                          #  [1, 0, 1],
                          #  [0, 1, 1]]
                          ]])
    print(true.shape)
    miou = Miou(num_classes, ignore=2)(pred, true)
    recall = Recall(num_classes, ignore=2)(pred, true)
    precision = Precision(num_classes, ignore=2)(pred, true)
    accuracy = Accuracy(num_classes, ignore=2)(pred, true)
    f1score = F1Score(num_classes, ignore=2)(pred, true)
    dice = DiceCoefficient(num_classes, ignore=2)(pred, true)
    aiu = AIU(num_classes, ignore=2)(pred, true)

    print("Mean IoU:", miou, type(miou))
    print("Recall:", recall, type(recall))
    print("Precision:", precision, type(precision))
    print("Accuracy:", accuracy, type(accuracy))
    print("F1Score:", f1score, type(f1score))
    print("DiceCoefficient:", dice, type(dice))
    print("AIU:", aiu, type(aiu))

    segmentation_index = SegmentationIndex(num_classes, ignore=2)
    segmentation_index.update(pred, true)
    print(segmentation_index)


import os
import string
import re
from pathlib import Path
import datetime

__all__=["Letterlabels", "Sorting"]

alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
uppercase_alphabet = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']

def Letterlabels(capital=False):
    """字母标签"""
    return uppercase_alphabet if capital else alphabet

class Sorting():
    """排序算法"""
    def insertion_sort(self, arr):
        """直接插入排序"""
        n = len(arr)
        for i in range(1, n):
            key = arr[i]
            j = i - 1
            while j >= 0 and arr[j] > key:
                arr[j + 1] = arr[j]
                j -= 1
            arr[j + 1] = key
            print(arr)

    def quick_sort(self, arr, low, high):
        """快速排序"""
        if low < high:
            pivot_index = self.partition(arr, low, high)
            print(arr)
            self.quick_sort(arr, low, pivot_index - 1)
            self.quick_sort(arr, pivot_index + 1, high)

    def partition(self, arr, low, high):
        pivot = arr[high]
        i = low - 1
        for j in range(low, high):
            if arr[j] <= pivot:
                i += 1
                arr[i], arr[j] = arr[j], arr[i]
        arr[i + 1], arr[high] = arr[high], arr[i + 1]
        return i + 1

    def bubble_sort(self, arr):
        """冒泡排序"""
        n = len(arr)
        for i in range(n - 1):
            for j in range(n - 1 - i):
                if arr[j] > arr[j + 1]:
                    arr[j], arr[j + 1] = arr[j + 1], arr[j]
            print(arr)

    def selection_sort(self, arr):
        """直接选择排序"""
        n = len(arr)
        for i in range(n - 1):
            min_index = i
            for j in range(i + 1, n):
                if arr[j] < arr[min_index]:
                    min_index = j
            arr[i], arr[min_index] = arr[min_index], arr[i]
            print(arr)



import cv2
import numpy as np
import pyzjr.Z as Z

__all__ = ["OverlayPng", "putBoxText", "cornerRect", "cvt2Center", "cvt2Corner", "AddText", "DrawPolygon", "DrawBboxPolygon"]

def OverlayPng(imgBack, imgFront, pos=(0, 0)):
    """
    叠加显示图片
    :param imgBack: 背景图像,无格式要求,3通道
    :param imgFront: png前置图片,读取方式必须使用 cv2.IMREAD_UNCHANGED = -1
    :param pos: 摆放位置
    :return: 叠加显示的图片
    """
    hf, wf, cf = imgFront.shape
    hb, wb, cb = imgBack.shape
    y_pos, x_pos = pos
    y_end = y_pos + hf
    x_end = x_pos + wf
    if y_end > hb:
        y_end = hb
    if x_end > wb:
        x_end = wb
    overlay_region = imgFront[:y_end - y_pos, :x_end - x_pos, :]
    overlay_mask = overlay_region[:, :, 3]
    overlay_alpha = overlay_mask.astype(float) / 255.0
    background_alpha = 1.0 - overlay_alpha
    result = overlay_region[:, :, :3] * overlay_alpha[..., np.newaxis] + imgBack[y_pos:y_end, x_pos:x_end, :3] * background_alpha[..., np.newaxis]
    imgBack[y_pos:y_end, x_pos:x_end, :3] = result

    return imgBack


def putBoxText(background_image, bbox, text, mode=1, rect=False, bboxcolor=(0, 255, 0), textcolor=(0, 0, 255),
               fontsize=1, thickness=2, bboxthickness=2,font=cv2.FONT_HERSHEY_SIMPLEX):
    """
    在给定的背景图像上绘制一个框，并在框的中心位置添加文本。获取文本的text_size，使文本居中。

    :param background_image: 背景图像，要在其上绘制框和文本的图像
    :param bbox: 框的边界框，表示为 [x1, y1, x2, y2]
    :param text: 要添加到框中心的文本
    :param mode: 模式0 表示框的中心位置，1 表示框的左上角
    :param rect: 是否绘制角标记框，导入的为 cornerRect 函数
    :param bboxcolor: 框的颜色，以 BGR 格式表示，例如 (0, 255, 0) 表示绿色
    :param textcolor: 文本的颜色，以 BGR 格式表示，例如 (0, 0, 255) 表示红色
    :param fontsize: 文本的字体大小
    :param thickness: 文本的线宽
    :param bboxthickness: 框的线宽，默认为 2
    :return: 绘制了框和文本的图像
    """
    text_x = text_y = None
    x1, y1, x2, y2 = bbox
    text_size, _ = cv2.getTextSize(text, font, fontsize, thickness)
    if rect:
        background_image = cornerRect(background_image, bbox)
    else:
        cv2.rectangle(background_image, (x1, y1), (x2, y2), bboxcolor, bboxthickness)
    if mode == 0:
        text_x = int((x1 + x2) / 2 - text_size[0] / 2)
        text_y = int((y1 + y2) / 2 + text_size[1] / 2)
    elif mode == 1:
        text_x = int(x1)
        text_y = int(y1 - text_size[1])
    cv2.putText(background_image, text, (text_x, text_y), font, fontsize, textcolor, thickness)


def cornerRect(drawimg, bbox, length=30, lthickness=5, rthickness=1, bboxcolor=Z.purple, colorCorner=Z.green):
    """
    在图像上绘制带有角标记的矩形框
    :param drawimg: 需要绘制的图像
    :param bbox: 边界框, 表示为 [x, y, x1, y1]
    :param length: 角标记的长度
    :param lthickness: 角标记的线条宽度
    :param rthickness: 矩形框的线条宽度
    :param bboxcolor: 矩形框的颜色
    :param colorCorner: 角标记的颜色
    :return: 绘制了角标记的图像副本
    """
    x, y, x1, y1 = bbox
    cv2.rectangle(drawimg, (x, y), (x1, y1), bboxcolor, rthickness)
    # 左上角  (x,y)
    cv2.line(drawimg, (x, y), (x + length, y), colorCorner, lthickness)
    cv2.line(drawimg, (x, y), (x, y + length), colorCorner, lthickness)
    # 右上角  (x1,y)
    cv2.line(drawimg, (x1, y), (x1 - length, y), colorCorner, lthickness)
    cv2.line(drawimg, (x1, y), (x1, y + length), colorCorner, lthickness)
    # 左下角  (x,y1)
    cv2.line(drawimg, (x, y1), (x + length, y1), colorCorner, lthickness)
    cv2.line(drawimg, (x, y1), (x, y1 - length), colorCorner, lthickness)
    # 右下角  (x1,y1)
    cv2.line(drawimg, (x1, y1), (x1 - length, y1), colorCorner, lthickness)
    cv2.line(drawimg, (x1, y1), (x1, y1 - length), colorCorner, lthickness)

    return drawimg

def cvt2Center(boxes):
    """
    Convert from (upper left, lower right) to (center, width, height).
    :param boxes: [x1, y1, x2, y2]
    :return: [cx, cy, w, h]
    """
    converted_boxes = []
    for box in boxes:
        x1, y1, x2, y2 = box[0], box[1], box[2], box[3]
        cx = (x1 + x2) / 2
        cy = (y1 + y2) / 2
        w = x2 - x1
        h = y2 - y1
        converted_box = [cx, cy, w, h]
        converted_boxes.append(converted_box)
    return converted_boxes

def cvt2Corner(boxes):
    """Convert from (center, width, height) to (upper left, lower right).
    :param boxes: [cx, cy, w, h]
    :return: [x1, y1, x2, y2]
    """
    converted_boxes = []
    for box in boxes:
        cx, cy, w, h = box[0], box[1], box[2], box[3]
        x1 = cx - 0.5 * w
        y1 = cy - 0.5 * h
        x2 = cx + 0.5 * w
        y2 = cy + 0.5 * h
        converted_box = [x1, y1, x2, y2]
        converted_boxes.append(converted_box)
    return converted_boxes

def AddText(img, text, x, y, color, thickness=3):
    """
    Usage Directions:
        Using the OpenCV method to add text to images
    Code Example:
    ------------
    >>> img = np.zeros((200, 400, 3), dtype=np.uint8)
    >>> AddText(img, "Hello, OpenCV!", 50, 100, (0, 255, 0), 2)
    """
    cv2.putText(
        img,
        text,
        (x, y),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=1,
        thickness=thickness,
        color=color,
        lineType=cv2.LINE_AA,
    )

def DrawPolygon(points, image, color):
    """
    Usage Directions:
        Draw a polygon onto an image using the given points and fill color.
    Code Example:
    ------------
    >>> image = np.ones((400, 400, 3), dtype=np.uint8)
    >>> polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    >>> result_image = DrawPolygon(polygon_points, image, (255, 0, 0))
    """
    points = np.array([points])
    points = points.astype(np.int32)
    image = cv2.fillPoly(image, points, color)
    return image

def DrawBboxPolygon(img, track_id, color, bbox, bboxcolor=(255, 0, 0)):
    """
    >>> bbox = np.array([100, 50, 300, 200])
    >>> result_image = DrawBboxPolygon(image, "Object 1", (0, 255, 0), bbox)
    """
    color = np.array(color)

    xmin, ymin, xmax, ymax = bbox.astype(np.int32).squeeze()

    bbox_h, bbox_w, _ = img[ymin:ymax, xmin:xmax].shape
    tiled_color = np.tile(color.reshape(1, 1, 3), (bbox_h, bbox_w, 1))
    img[ymin:ymax, xmin:xmax, :] = (img[ymin:ymax, xmin:xmax, :] + tiled_color) / 2.0

    white = (255, 255, 255)
    plot_x = xmin + 10
    plot_y = ymin + 25

    AddText(
        img,
        str(track_id),
        plot_x,
        plot_y,
        color=white,
        thickness=5,
    )
    img = cv2.rectangle(img, (xmin, ymin), (xmax, ymax), color=bboxcolor)
    return img


"""
           _oo0oo_
          o8888888o
          88" . "88
          (| -_- |)
          0\  =  /0
        ___/`---'\___
      .' \\|     |// '.
     / \\|||  :  |||// \
    / _||||| -:- |||||_ \
   |   | \\\  -  /'| |   |
   | \_|  ''\---/''  |_/ |
   \  .-\__  '-'  __/-.  /
 ====='-._____'-----'___.-'=====
"""


import os
import numpy as np
import pandas as pd
import matplotlib
import torch
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import scipy.signal as signal
from torch.utils.tensorboard import SummaryWriter

__all__ = ["LossHistory", "ErrorRateMonitor", "LossMonitor"]

class LossHistory():
    def __init__(self, log_dir, model, input_shape):
        self.log_dir = log_dir
        self.losses = []
        self.val_loss = []

        os.makedirs(self.log_dir,exist_ok=True)
        self.writer = SummaryWriter(self.log_dir)
        try:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
            dummy_input = torch.randn(2, 3, input_shape[0], input_shape[1]).to(device)
            self.writer.add_graph(model, dummy_input)
        except:
            pass

    def append_loss(self, epoch, loss, val_loss):
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.losses.append(loss)
        self.val_loss.append(val_loss)

        with open(os.path.join(self.log_dir, "epoch_loss.txt"), 'a') as f:
            f.write(f"Epoch {epoch}: Train: {loss}, Validation: {val_loss}")
            f.write("\n")

        self.writer.add_scalar('loss', loss, epoch)
        self.writer.add_scalar('val_loss', val_loss, epoch)
        self.loss_plot()

    def loss_plot(self):
        iters = range(len(self.losses))

        plt.figure()
        plt.plot(iters, self.losses, 'red', linewidth = 2, label='train loss')
        plt.plot(iters, self.val_loss, 'coral', linewidth = 2, label='val loss')
        try:
            if len(self.losses) < 25:
                num = 5
            else:
                num = 15

            plt.plot(iters, signal.savgol_filter(self.losses, num, 3), 'green', linestyle = '--', linewidth = 2, label='smooth train loss')
            plt.plot(iters, signal.savgol_filter(self.val_loss, num, 3), '#8B4513', linestyle = '--', linewidth = 2, label='smooth val loss')
        except:
            pass

        plt.grid(True)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.legend(loc="upper right")

        plt.savefig(os.path.join(self.log_dir, "epoch_loss.png"))

        plt.cla()
        plt.close("all")

    def close(self):
        self.writer.close()

class ErrorRateMonitor:
    """
    错误率监视器,传入train和val的准确率,记录错误率
    """
    def __init__(self, log_dir):
        self.save_path = os.path.join(log_dir, "Error_rate.png")
        self.fig, self.ax = plt.subplots()
        self.train_error_rates = []
        self.val_error_rates = []
        self.acc_log_path = os.path.join(log_dir, "acc_log.txt")
        self.ax.set_xlabel('Epoch')
        self.ax.set_ylabel('Error_rate')

    def append_acc(self, epoch, train_acc, val_acc):
        train_error_rate = 1 - train_acc
        val_error_rate = 1 - val_acc

        self.train_error_rates.append(train_error_rate)
        self.val_error_rates.append(val_error_rate)

        plt.title(f'Epoch {epoch}')
        with open(self.acc_log_path, 'a') as acc_file:
            acc_file.write(
                f"Epoch {epoch}: Train Accuracy: {train_acc:.4f}, Validation Accuracy: {val_acc:.4f}\n")
        self.error_plot(epoch)

    def error_plot(self, epoch):
        iters = np.arange(1, epoch + 1)

        self.ax.clear()
        self.ax.plot(iters, self.train_error_rates[:len(iters)], 'red', linewidth=2, label='train error')
        self.ax.plot(iters, self.val_error_rates[:len(iters)], 'coral', linewidth=2, label='val error')
        try:
            if len(self.train_error_rates) < 25:
                num = 5
            else:
                num = 15

            self.ax.plot(iters, signal.savgol_filter(self.train_error_rates[:len(iters)], num, 3), 'green',
                         linestyle='--',
                         linewidth=2, label='smooth train error')
            self.ax.plot(iters, signal.savgol_filter(self.val_error_rates[:len(iters)], num, 3), '#8B4513',
                         linestyle='--',
                         linewidth=2, label='smooth val error')
        except:
            pass

        self.ax.grid(True)
        self.ax.legend(loc="upper right")

        self.fig.savefig(self.save_path)

    def close(self):
        plt.close("all")


class LossMonitor:
    def __init__(self, epochs, metric='train_loss', mode='min',
                 save_path='history_loss.png', figsize=(8, 6)):
        self.figsize = figsize
        self.save_path = save_path
        self.metric_name = metric
        self.metric_mode = mode
        self.epochs = epochs
        self.history = {}
        self.step, self.epoch = 0, 0

    def start(self):
        print('\nView dynamic loss/metric plot: \n' + os.path.abspath(self.save_path))
        x_bounds = [0, min(10, self.epochs)]
        title = f'best {self.metric_name} = ?'
        self.update_graph(title=title, x_bounds=x_bounds)

    def log_epoch(self, info):
        self.epoch += 1
        info['epoch'] = self.epoch
        for name, metric in info.items():
            self.history[name] = self.history.get(name, []) + [metric]
        dfhistory = pd.DataFrame(self.history)
        n = len(dfhistory)
        x_bounds = [dfhistory['epoch'].min(), min(10 + (n // 10) * 10, self.epochs)]
        title = self.get_title()
        self.step, self.batchs = 0, self.step
        self.update_graph(title=title, x_bounds=x_bounds)

    def end(self):
        title = self.get_title()
        self.update_graph(title=title)
        self.dfhistory = pd.DataFrame(self.history)
        return self.dfhistory

    def head(self):
        best_epoch, best_score = self.get_best_score()
        print(f"Best {self.metric_name}: {best_score:.4f} at epoch {best_epoch}")
        return self.dfhistory.head()

    def get_best_score(self):
        dfhistory = pd.DataFrame(self.history)
        arr_scores = dfhistory[self.metric_name]
        best_score = np.max(arr_scores) if self.metric_mode == "max" else np.min(arr_scores)
        best_epoch = dfhistory.loc[arr_scores == best_score, 'epoch'].tolist()[0]
        return (best_epoch, best_score)

    def get_title(self):
        best_epoch, best_score = self.get_best_score()
        title = f'best {self.metric_name}={best_score:.4f} (@epoch {best_epoch})'
        return title

    def update_graph(self, title=None, x_bounds=None, y_bounds=None):
        if not hasattr(self, 'graph_fig'):
            self.fig, self.ax = plt.subplots(1, figsize=self.figsize)
        self.ax.clear()

        dfhistory = pd.DataFrame(self.history)
        epochs = dfhistory['epoch'] if 'epoch' in dfhistory.columns else []

        metric_name = self.metric_name.replace('val_', '').replace('train_', '')

        m1 = "train_" + metric_name
        if m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            self.ax.plot(epochs, train_metrics, 'bo--', label=m1, clip_on=False)

        m2 = 'val_' + metric_name
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.ax.plot(epochs, val_metrics, 'co-', label=m2, clip_on=False)

        if metric_name in dfhistory.columns:
            metric_values = dfhistory[metric_name]
            self.ax.plot(epochs, metric_values, 'co-', label=self.metric_name, clip_on=False)

        self.ax.set_xlabel("epoch")
        self.ax.set_ylabel(metric_name)

        if title:
            self.ax.set_title(title)

        if m1 in dfhistory.columns or m2 in dfhistory.columns or self.metric_name in dfhistory.columns:
            self.ax.legend(loc='best')

        if len(epochs) > 0:
            best_epoch, best_score = self.get_best_score()
            self.ax.plot(best_epoch, best_score, 'r*', markersize=15, clip_on=False)

        if x_bounds is not None: self.ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.ax.set_ylim(*y_bounds)
        self.fig.savefig(self.save_path)
        plt.close()



if __name__=='__main__':
    import random
    epochs = 20
    vlog = LossMonitor(epochs=epochs, metric='train_loss', mode='min')

    for epoch in range(1, epochs + 1):
        train_loss = random.uniform(0.5, 2.0)
        val_loss = random.uniform(0.4, 1.5)
        vlog.log_epoch({'train_loss': train_loss, 'val_loss': val_loss})

    df_history = vlog.end()
    print(vlog.head())



"""
论文原址： <https://arxiv.org/pdf/1512.03385.pdf>
Deep Residual Learning for Image Recognition
resnet50, resnet101, resnet152
"""
import torch
import torch.nn as nn

__all__ = ["ResNet", 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']

class ConvBNReLU(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        self.bn = nn.BatchNorm2d(num_features=out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

class ConvBN(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1):
        super(ConvBN, self).__init__()
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        self.bn = nn.BatchNorm2d(num_features=out_channels)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x

def conv3x3(in_channels, out_channels, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding:捕捉局部特征和空间相关性，学习更复杂的特征和抽象表示"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv1x1(in_channels, out_channels, stride=1):
    """1x1 convolution:实现降维或升维，调整通道数和执行通道间的线性变换"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

class BasicBlock(nn.Module):
    expansion = 1
    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.convbnrelu1 = ConvBNReLU(in_channels, out_channels, kernel_size=3, stride=stride)
        self.convbn1 = ConvBN(out_channels, out_channels, kernel_size=3)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(in_channels, out_channels * self.expansion, self.stride),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

    def forward(self, x):
        residual = x
        out = self.convbnrelu1(x)
        out = self.convbn1(out)

        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4
    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        groups = 1
        base_width = 64
        dilation = 1

        width = int(out_channels * (base_width / 64.)) * groups   # wide = out_channels
        self.conv1 = conv1x1(in_channels, width)       # 降维通道数
        self.bn1 = nn.BatchNorm2d(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = conv1x1(width, out_channels * self.expansion)   # 升维通道数
        self.bn3 = nn.BatchNorm2d(out_channels * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(in_channels, out_channels * self.expansion, self.stride),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out

class ResNet(nn.Module):
    def __init__(self, block, layers, num_classes=1000, zero_init_residual=False,
                 groups=1, width_per_group=64):
        super(ResNet, self).__init__()

        self.inplanes = 64
        self.dilation = 1
        replace_stride_with_dilation = [False, False, False]
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        downsample = False
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = True

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):    # 添加几个残差块, 跟resnet的结构有关
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)    # torch.Size([2, 64, 56, 56])
        x = self.layer1(x)     # torch.Size([2, 256, 56, 56])
        x = self.layer2(x)     # torch.Size([2, 512, 56, 56])
        x = self.layer3(x)     # torch.Size([2, 1024, 28, 28])
        x = self.layer4(x)     # torch.Size([2, 2048, 14, 14])

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

def resnet18(num_classes, **kwargs):
    return ResNet(BasicBlock, [2, 2, 2, 2], num_classes=num_classes, **kwargs)

def resnet34(num_classes, **kwargs):
    return ResNet(BasicBlock, [3, 4, 6, 3], num_classes=num_classes, **kwargs)

def resnet50(num_classes, **kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], num_classes=num_classes, **kwargs)

def resnet101(num_classes, **kwargs):
    return ResNet(Bottleneck, [3, 4, 23, 3], num_classes=num_classes, **kwargs)

def resnet152(num_classes, **kwargs):
    return ResNet(Bottleneck, [3, 8, 36, 3], num_classes=num_classes, **kwargs)

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = resnet50(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # 50 Total params: 23,516,228